# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import itertools
import logging
import sys
import builtins
from contextlib import nullcontext

from pyomo.common.collections import MutableMapping
from pyomo.common.errors import TemplateExpressionError
from pyomo.common.gc_manager import PauseGC
from pyomo.core.expr.base import ExpressionBase, ExpressionArgs_Mixin, NPV_Mixin
from pyomo.core.expr.logical_expr import BooleanExpression
from pyomo.core.expr.numeric_expr import (
    ARG_TYPE,
    NumericExpression,
    Numeric_NPV_Mixin,
    SumExpression,
    mutable_expression,
    register_arg_type,
    _balanced_parens,
)
from pyomo.core.expr.numvalue import (
    NumericValue,
    native_types,
    nonpyomo_leaf_types,
    as_numeric,
    value,
    is_constant,
)
from pyomo.core.expr.relational_expr import tuple_to_relational_expr
from pyomo.core.expr.visitor import (
    ExpressionReplacementVisitor,
    StreamBasedExpressionVisitor,
    expression_to_string,
    _ToStringVisitor,
)

logger = logging.getLogger(__name__)


class _NotSpecified:
    pass


class GetItemExpression(ExpressionBase):
    """
    Expression to call :func:`__getitem__` on the base object.
    """

    __slots__ = ()
    PRECEDENCE = 1

    def __new__(cls, args=()):
        if cls is not GetItemExpression:
            return super().__new__(cls)
        npv_args = not any(
            hasattr(arg, 'is_potentially_variable') and arg.is_potentially_variable()
            for arg in args
        )
        try:
            component = _reduce_template_to_component(args[0])
            cdata = component._ComponentDataClass(component)
            if cdata.is_numeric_type():
                if npv_args and not cdata.is_potentially_variable():
                    return super().__new__(NPV_Numeric_GetItemExpression)
                else:
                    return super().__new__(Numeric_GetItemExpression)
            if cdata.is_logical_type():
                if npv_args and not cdata.is_potentially_variable():
                    return super().__new__(NPV_Boolean_GetItemExpression)
                else:
                    return super().__new__(Boolean_GetItemExpression)
        except (AttributeError, TypeError):
            # TypeError: error reducing to a component (usually due to
            #     unbounded domain on a Var used in a GetItemExpression)
            # AttributeError: resolved component did not support the
            #     PyomoObject API
            pass
        if npv_args:
            return super().__new__(NPV_Structural_GetItemExpression)
        else:
            return super().__new__(Structural_GetItemExpression)

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):
            raise AttributeError()
        return GetAttrExpression((self, attr))

    def __iter__(self):
        return iter(value(self))

    def __len__(self):
        return len(value(self))

    def getname(self, *args, **kwds):
        return self._args_[0].getname(*args, **kwds)

    def nargs(self):
        return len(self._args_)

    def _is_fixed(self, values):
        if not all(values[1:]):
            return False
        _true = lambda: True
        return all(getattr(x, 'is_fixed', _true)() for x in values[0].values())

    def _to_string(self, values, verbose, smap):
        values = tuple(_[1:-1] if _[0] == '(' and _[-1] == ')' else _ for _ in values)
        if verbose:
            return "getitem(%s, %s)" % (values[0], ', '.join(values[1:]))
        return "%s[%s]" % (values[0], ','.join(values[1:]))

    def _resolve_template(self, args):
        return args[0].__getitem__(args[1:])

    def _apply_operation(self, result):
        return result[0].__getitem__(result[1:])


class Numeric_GetItemExpression(GetItemExpression, NumericExpression):
    __slots__ = ()

    def nargs(self):
        return len(self._args_)

    def _compute_polynomial_degree(self, result):
        if any(x != 0 for x in result[1:]):
            return None
        ans = 0
        for x in result[0].values():
            if x.__class__ in nonpyomo_leaf_types or not hasattr(
                x, 'polynomial_degree'
            ):
                continue
            tmp = x.polynomial_degree()
            if tmp is None:
                return None
            elif tmp > ans:
                ans = tmp
        return ans


class NPV_Numeric_GetItemExpression(Numeric_NPV_Mixin, Numeric_GetItemExpression):
    __slots__ = ()


class Boolean_GetItemExpression(GetItemExpression, BooleanExpression):
    __slots__ = ()


class NPV_Boolean_GetItemExpression(NPV_Mixin, Boolean_GetItemExpression):
    __slots__ = ()


class Structural_GetItemExpression(ExpressionArgs_Mixin, GetItemExpression):
    __slots__ = ()


class NPV_Structural_GetItemExpression(NPV_Mixin, Structural_GetItemExpression):
    __slots__ = ()


class GetAttrExpression(ExpressionBase):
    """
    Expression to call :func:`__getattr__` on the base object.
    """

    __slots__ = ()
    PRECEDENCE = 1

    def __new__(cls, args=()):
        if cls is not GetAttrExpression:
            return super().__new__(cls)
        # Ironically, we need to actually create this object in order to
        # determine what the class for this object should be.
        if args[0].is_potentially_variable():
            self = Structural_GetAttrExpression(args)
        else:
            self = NPV_Structural_GetAttrExpression(args)
        try:
            attr = _reduce_template_to_component(self)
            if attr.is_numeric_type():
                if attr.is_potentially_variable() or self.is_potentially_variable():
                    return super().__new__(Numeric_GetAttrExpression)
                else:
                    return super().__new__(NPV_Numeric_GetAttrExpression)
            elif attr.is_logical_type():
                if attr.is_potentially_variable() or self.is_potentially_variable():
                    return super().__new__(Boolean_GetAttrExpression)
                else:
                    return super().__new__(NPV_Boolean_GetAttrExpression)
        except (AttributeError, TypeError):
            # TypeError: error reducing to a component (usually due to
            #     unbounded domain on a Var used in a GetItemExpression)
            # AttributeError: resolved component did not support the
            #     PyomoObject API
            pass
        return self

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):
            raise AttributeError()
        return GetAttrExpression((self, attr))

    def __getitem__(self, *idx):
        return GetItemExpression((self,) + idx)

    def __iter__(self):
        return iter(value(self))

    def __len__(self):
        return len(value(self))

    def __call__(self, *args, **kwargs):
        """
        Return the value of this object.
        """
        # Backwards compatibility with __call__(exception):
        #
        # TODO: deprecate (then remove) evaluating expressions by
        # "calling" them.
        #
        # [ESJ 3/25/25]: Note that since this always calls the ExpressionBase
        # implementation of __call__ if 'exception' is specified, we need not
        # check the type of the exception arg here--it will get checked in the
        # base class.
        try:
            if not args:
                if not kwargs:
                    return super().__call__()
                elif len(kwargs) == 1 and 'exception' in kwargs:
                    return super().__call__(**kwargs)
            elif (
                not kwargs and len(args) == 1 and (args[0] is True or args[0] is False)
            ):
                return super().__call__(*args)
        except TemplateExpressionError:
            pass
        # Note: the only time we will implicitly create a CallExpression
        # node is directly after a GetAttrExpression: that is, someone
        # got the attribute (method) and is now calling it.
        # Implementing the auto-generation of CallExpression in other
        # contexts is likely to be confounded with evaluating expressions.
        return CallExpression((self,) + args, kwargs)

    def getname(self, *args, **kwds):
        return 'getattr'

    def nargs(self):
        return 2

    def _apply_operation(self, result):
        obj, attr = result
        return getattr(obj, attr)

    def _to_string(self, values, verbose, smap):
        assert len(values) == 2
        if verbose:
            return "getattr(%s, %s)" % tuple(values)
        # Note that the string argument for getattr comes quoted, so we
        # need to remove the quotes.
        attr = values[1]
        if attr[0] in '\"\'' and attr[0] == attr[-1]:
            attr = attr[1:-1]
        return "%s.%s" % (values[0], attr)

    def _resolve_template(self, args):
        return getattr(*args)


class Numeric_GetAttrExpression(GetAttrExpression, NumericExpression):
    __slots__ = ()

    def _compute_polynomial_degree(self, result):
        if result[1] != 0:
            return None
        return result[0]


class NPV_Numeric_GetAttrExpression(Numeric_NPV_Mixin, Numeric_GetAttrExpression):
    __slots__ = ()


class Boolean_GetAttrExpression(GetAttrExpression, BooleanExpression):
    __slots__ = ()


class NPV_Boolean_GetAttrExpression(NPV_Mixin, Boolean_GetAttrExpression):
    __slots__ = ()


class Structural_GetAttrExpression(ExpressionArgs_Mixin, GetAttrExpression):
    __slots__ = ()


class NPV_Structural_GetAttrExpression(NPV_Mixin, Structural_GetAttrExpression):
    __slots__ = ()


class CallExpression(NumericExpression):
    """
    Expression to call :func:`__call__` on the base object.
    """

    __slots__ = ('_kwds',)
    PRECEDENCE = None

    def __init__(self, args, kwargs):
        self._args_ = tuple(args) + tuple(kwargs.values())
        self._kwds = tuple(kwargs.keys())

    def nargs(self):
        return len(self._args_)

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):
            raise AttributeError()
        return GetAttrExpression((self, attr))

    def __getitem__(self, *idx):
        return GetItemExpression((self,) + idx)

    def __iter__(self):
        return iter(value(self))

    def __len__(self):
        return len(value(self))

    def getname(self, *args, **kwds):
        return 'call'

    def _compute_polynomial_degree(self, result):
        return None

    def _apply_operation(self, result):
        na = len(self._args_) - len(self._kwds)
        return result[0](*result[1:na], **dict(zip(self._kwds, result[na:])))

    def _to_string(self, values, verbose, smap):
        na = len(self._args_) - len(self._kwds)
        args = ', '.join(values[1:na])
        if self._kwds:
            if na > 1:
                args += ', '
            args += ', '.join(
                f'{key}={val}' for key, val in zip(self._kwds, values[na:])
            )
        if verbose:
            return f"call({values[0]}, {args})"
        return f"{values[0]}({args})"

    def _resolve_template(self, args):
        return self._apply_operation(args)


class _TemplateSumExpression_argList:
    """A virtual list to represent the expanded SumExpression args

    This class implements a "virtual args list" for
    TemplateSumExpressions without actually generating the expanded
    expression.  It can be accessed either in "one-pass" without
    generating a list of template argument values (more efficient), or
    as a random-access list (where it will have to create the full list
    of argument values (less efficient).

    The instance can be used as a context manager to both lock the
    IndexTemplate values within this context and to restore their original
    values upon exit.

    It is (intentionally) not iterable.

    """

    def __init__(self, TSE):
        self._tse = TSE
        self._i = 0
        self._init_vals = None
        self._iter = self._get_iter()
        self._lock = None

    def __len__(self):
        return self._tse.nargs()

    def __getitem__(self, i):
        if self._i == i:
            self._set_iter_vals(next(self._iter))
            self._i += 1
        elif self._i is not None:
            # Switch to random-access mode.  If we have already
            # retrieved one of the indices, then we need to regenerate
            # the iterator from scratch.
            self._iter = list(self._get_iter() if self._i else self._iter)
            self._set_iter_vals(self._iter[i])
        else:
            self._set_iter_vals(self._iter[i])
        return self._tse._local_args_[0]

    def __enter__(self):
        self._lock = self
        self._lock_iters()

    def __exit__(self, exc_type, exc_value, tb):
        self._unlock_iters()
        self._lock = None

    def _get_iter(self):
        # Note: by definition, all _set pointers within an itergroup
        # point to the same Set
        _sets = tuple(iterGroup[0]._set for iterGroup in self._tse._iters)
        return itertools.product(*_sets)

    def _lock_iters(self):
        self._init_vals = tuple(
            tuple(it.lock(self._lock) for it in iterGroup)
            for iterGroup in self._tse._iters
        )

    def _unlock_iters(self):
        self._set_iter_vals(self._init_vals)
        for iterGroup in self._tse._iters:
            for it in iterGroup:
                it.unlock(self._lock)

    def _set_iter_vals(self, val):
        for i, iterGroup in enumerate(self._tse._iters):
            if len(iterGroup) == 1:
                iterGroup[0].set_value(val[i], self._lock)
            else:
                for j, v in enumerate(val[i]):
                    iterGroup[j].set_value(v, self._lock)


class TemplateSumExpression(NumericExpression):
    """
    Expression to represent an unexpanded sum over one or more sets.
    """

    __slots__ = ('_iters', '_local_args_')
    PRECEDENCE = 1

    def __init__(self, args, _iters):
        assert len(args) == 1
        self._args_ = args
        self._iters = _iters

    def nargs(self):
        # Note: by definition, all _set pointers within an itergroup
        # point to the same Set
        ans = 1
        for iterGroup in self._iters:
            ans *= len(iterGroup[0]._set)
        return ans

    @property
    def args(self):
        return _TemplateSumExpression_argList(self)

    @property
    def _args_(self):
        return _TemplateSumExpression_argList(self)

    @_args_.setter
    def _args_(self, args):
        self._local_args_ = args

    def template_args(self):
        ans = list(self._local_args_)
        for itergroup in self._iters:
            ans.append(itergroup[0]._set)
        return tuple(ans)

    def template_iters(self):
        return self._iters

    def create_node_with_local_data(self, args):
        return self.__class__(args, self._iters)

    def getname(self, *args, **kwds):
        return "SUM"

    def is_potentially_variable(self):
        if any(
            arg.is_potentially_variable()
            for arg in self._local_args_
            if arg.__class__ not in nonpyomo_leaf_types
        ):
            return True
        return False

    def _is_fixed(self, values):
        return all(values)

    def _compute_polynomial_degree(self, result):
        if None in result:
            return None
        return result[0]

    def _apply_operation(self, result):
        return sum(result)

    def to_string(self, verbose=None, smap=None):
        ans = ''
        assert len(self._local_args_) == 1
        val = expression_to_string(self._local_args_[0], verbose=verbose, smap=smap)
        if val[0] == '(' and val[-1] == ')' and _balanced_parens(val[1:-1]):
            val = val[1:-1]
        iterStrGenerator = (
            (
                ', '.join(
                    (smap.getSymbol(i) if smap is not None else str(i))
                    for i in iterGroup
                ),
                (
                    iterGroup[0]._set.to_string(verbose=verbose, smap=smap)
                    if hasattr(iterGroup[0]._set, 'to_string')
                    else (
                        smap.getSymbol(iterGroup[0]._set)
                        if smap is not None
                        else str(iterGroup[0]._set)
                    )
                ),
            )
            for iterGroup in self._iters
        )
        if verbose:
            iterStr = ', '.join('iter(%s, %s)' % x for x in iterStrGenerator)
            return 'templatesum(%s, %s)' % (val, iterStr)
        else:
            iterStr = ' '.join('for %s in %s' % x for x in iterStrGenerator)
            return 'SUM(%s %s)' % (val, iterStr)

    def _resolve_template(self, args):
        with mutable_expression() as e:
            for arg in args:
                e += arg
        if e.nargs() > 1:
            return e
        elif not e.nargs():
            return 0
        else:
            return e.arg(0)


# FIXME: This is a hack to get certain complex cases to print without error
_ToStringVisitor._leaf_node_types.add(TemplateSumExpression)


class IndexTemplate(NumericValue):
    """A "placeholder" for an index value in template expressions.

    This class is a placeholder for an index value within a template
    expression.  That is, given the expression template for "m.x[i]",
    where `m.z` is indexed by `m.I`, the expression tree becomes:

    _GetItem:
       - m.x
       - IndexTemplate(_set=m.I, _value=None)

    Constructor Arguments:
       _set: the Set from which this IndexTemplate can take values
    """

    __slots__ = ('_set', '_value', '_index', '_id', '_group', '_lock')

    def __init__(self, _set, index=0, _id=None, _group=None):
        self._set = _set
        self._value = _NotSpecified
        self._index = index
        self._id = _id
        self._group = _group
        self._lock = None

    def __deepcopy__(self, memo):
        # Because we leverage deepcopy for expression/component cloning,
        # we need to see if this is a Component.clone() operation and
        # *not* copy the template.
        #
        # TODO: JDS: We should consider converting the IndexTemplate to
        # a proper Component: that way it could leverage the normal
        # logic of using the parent_block scope to dictate the behavior
        # of deepcopy.
        if '__block_scope__' in memo:
            memo[id(self)] = self
            return self
        #
        # "Normal" deepcopying outside the context of pyomo.
        #
        return super().__deepcopy__(memo)

    # Note: because NONE of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    def __call__(self, exception=True):
        """
        Return the value of this object.
        """
        if self._value is _NotSpecified:
            if exception:
                raise TemplateExpressionError(
                    self, "Evaluating uninitialized IndexTemplate (%s)" % (self,)
                )
            return None
        else:
            return self._value

    def _resolve_template(self, args):
        assert not args
        return self()

    def is_fixed(self):
        """
        Returns True because this value is fixed.
        """
        return True

    def is_potentially_variable(self):
        """Returns False because index values cannot be variables.

        The IndexTemplate represents a placeholder for an index value
        for an IndexedComponent, and at the moment, Pyomo does not
        support variable indirection.
        """
        return False

    def __str__(self):
        return self.getname()

    def getname(self, fully_qualified=False, name_buffer=None, relative_to=None):
        if self._id is not None:
            return "_%s" % (self._id,)

        _set_name = self._set.getname(fully_qualified, name_buffer, relative_to)
        if self._index is not None and self._set.dimen != 1:
            _set_name += "(%s)" % (self._index,)
        return "{" + _set_name + "}"

    def set_value(self, values=_NotSpecified, lock=None):
        # It might be nice to check if the value is valid for the base
        # set, but things are tricky when the base set is not dimension
        # 1.  So, for the time being, we will just "trust" the user.
        # After all, the actual Set will raise exceptions if the value
        # is not present.
        if lock is not self._lock:
            raise RuntimeError(
                "The IndexTemplate %s is currently locked by %s and "
                "cannot be set through lock %s" % (self, self._lock, lock)
            )
        if values is _NotSpecified:
            self._value = _NotSpecified
            return
        if type(values) is not tuple:
            values = (values,)
        if self._index is not None:
            if len(values) == 1:
                self._value = values[0]
            else:
                self._value = values[self._index]
        else:
            self._value = values

    def lock(self, lock):
        assert self._lock is None
        self._lock = lock
        return self._value

    def unlock(self, lock):
        assert self._lock is lock
        self._lock = None


# Instead of special-casing _categorize_arg_type for this class, we
# will directly register that it should be treated as an NPV arg
register_arg_type(IndexTemplate, ARG_TYPE.NPV)


class _TemplateResolver(StreamBasedExpressionVisitor):
    def beforeChild(self, node, child, child_idx):
        # Efficiency: do not descend into leaf nodes.
        if type(child) in native_types:
            return False, child
        elif not child.is_expression_type():
            if hasattr(child, '_resolve_template'):
                return False, child._resolve_template(())
            return False, child
        else:
            return True, None

    def exitNode(self, node, args):
        if hasattr(node, '_resolve_template'):
            return node._resolve_template(args)
        if len(args) == node.nargs() and all(a is b for a, b in zip(node.args, args)):
            return node
        if all(map(is_constant, args)):
            return node._apply_operation(args)
        else:
            return node.create_node_with_local_data(args)

    def initializeWalker(self, expr):
        return self.beforeChild(None, expr, None)


def resolve_template(expr):
    """Resolve a template into a concrete expression

    This takes a template expression and returns the concrete equivalent
    by substituting the current values of all IndexTemplate objects and
    resolving (evaluating and removing) all GetItemExpression,
    GetAttrExpression, and TemplateSumExpression expression nodes.

    """
    if resolve_template.visitor is None:
        resolve_template.visitor = _TemplateResolver()
    return resolve_template.visitor.walk_expression(expr)


resolve_template.visitor = None


class _wildcard_info:
    __slots__ = ('iter', 'source', 'value', 'original_value', 'objects')

    def __init__(self, src, obj):
        self.source = src
        self.original_value = obj._value
        self.objects = [obj]
        self.reset()
        if self.original_value in (None, _NotSpecified):
            self.advance()

    def advance(self):
        with _TemplateIterManager.pause():
            self.value = next(self.iter)
        for obj in self.objects:
            obj.set_value(self.value)

    def reset(self):
        # Because we want to actually iterate over the underlying
        # template expression, we will temporarily pause our overrides
        # of sum() and the set iters
        with _TemplateIterManager.pause():
            self.iter = iter(self.source)

    def restore(self):
        for obj in self.objects:
            obj.set_value(self.original_value)


def _reduce_template_to_component(expr):
    """Resolve a template into a concrete component

    This takes a template expression and returns the concrete equivalent
    by substituting the current values of all IndexTemplate objects and
    resolving (evaluating and removing) all GetItemExpression,
    GetAttrExpression, and TemplateSumExpression expression nodes.

    """
    import pyomo.core.base.set

    # wildcards holds lists of
    #   [iterator, source, value, orig_value, object0, ...]
    # 'iterator' iterates over 'source' to provide 'value's for each of
    # the 1 or more 'objects'.  Objects can be IndexTemplate objects or
    # (discrete) Variables
    wildcards = []
    wildcard_groups = {}
    level = -1

    def beforeChild(node, child, child_idx):
        # Efficiency: do not descend into leaf nodes.
        if type(child) in native_types:
            return False, child
        elif not child.is_expression_type():
            if hasattr(child, '_resolve_template'):
                try:
                    ans = child._resolve_template(())
                except TemplateExpressionError:
                    # We are attempting "loose" template resolution: for
                    # every unset IndexTemplate, search the underlying
                    # set to find *any* valid match.
                    if child._group not in wildcard_groups:
                        wildcard_groups[child._group] = len(wildcards)
                        info = _wildcard_info(child._set, child)
                        wildcards.append(info)
                    else:
                        info = wildcards[wildcard_groups[child._group]]
                        info.objects.append(child)
                        child.set_value(info.value)
                    ans = child._resolve_template(())
                return False, ans
            if child.is_variable_type():
                from pyomo.core.base.set import RangeSet

                if child.domain.isdiscrete():
                    domain = child.domain
                    bounds = child.bounds
                    if bounds != (None, None):
                        try:
                            bounds = pyomo.core.base.set.RangeSet(*bounds, 0)
                            domain = domain & bounds
                        except:
                            pass
                    info = _wildcard_info(domain, child)
                    wildcards.append(info)
                return False, value(child)
            return False, child
        else:
            return True, None

    def exitNode(node, args):
        if hasattr(node, '_resolve_template'):
            return node._resolve_template(args)
        if len(args) == node.nargs() and all(a is b for a, b in zip(node.args, args)):
            return node
        if all(map(is_constant, args)):
            return node._apply_operation(args)
        else:
            return node.create_node_with_local_data(args)

    walker = StreamBasedExpressionVisitor(
        initializeWalker=lambda x: beforeChild(None, x, None),
        beforeChild=beforeChild,
        exitNode=exitNode,
    )
    while 1:
        try:
            with _TemplateIterManager.pause():
                ans = walker.walk_expression(expr)
            break
        except (KeyError, AttributeError):
            # We are attempting "loose" template resolution: for every
            # unset IndexTemplate, search the underlying set to find
            # *any* valid match.
            level = len(wildcards) - 1
            while level >= 0:
                info = wildcards[level]
                try:
                    info.advance()
                    break
                except StopIteration:
                    # Because we want to actually iterate over the
                    # underlying template expression, we will
                    # temporarily pause our overrides of sum() and the
                    # set iters
                    info.reset()
                    info.advance()
                    level -= 1
            if level < 0:
                for info in wildcards:
                    info.restore()
                raise
    for info in wildcards:
        info.restore()
    return ans


class ReplaceTemplateExpression(ExpressionReplacementVisitor):
    template_types = {
        IndexTemplate,
        GetItemExpression,
        Numeric_GetItemExpression,
        NPV_Numeric_GetItemExpression,
        Boolean_GetItemExpression,
        NPV_Boolean_GetItemExpression,
    }

    def __init__(self, substituter, *args, **kwargs):
        kwargs.setdefault('remove_named_expressions', True)
        super().__init__(**kwargs)
        self.substituter = substituter
        self.substituter_args = args

    def beforeChild(self, node, child, child_idx):
        if type(child) in ReplaceTemplateExpression.template_types:
            return False, self.substituter(child, *self.substituter_args)
        return super().beforeChild(node, child, child_idx)


def substitute_template_expression(expr, substituter, *args, **kwargs):
    r"""Substitute IndexTemplates in an expression tree.

    This is a general utility function for walking the expression tree
    and substituting all occurrences of IndexTemplate and
    GetItemExpression nodes.

    Parameters
    ----------
    expr : NumericExpression
        the source template expression

    substituter: Callable
        method taking ``(expression, *args)`` and returning the new object

    \*args:
        positional arguments passed directly to the substituter

    Returns
    -------
    NumericExpression :
        a new expression tree with all substitutions done

    """
    visitor = ReplaceTemplateExpression(substituter, *args, **kwargs)
    return visitor.walk_expression(expr)


class _GetItemIndexer:
    # Note that this class makes the assumption that only one template
    # ever appears in an expression for a single index

    def __init__(self, expr):
        self._base = expr.arg(0)
        self._args = []
        _hash = [id(self._base)]
        for x in expr.args[1:]:
            try:
                logging.disable(logging.CRITICAL)
                val = value(x)
                self._args.append(val)
                _hash.append(val)
            except TemplateExpressionError as e:
                if x is not e.template:
                    raise TypeError(
                        "Cannot use the param substituter with expression "
                        "templates\nwhere the component index has the "
                        "IndexTemplate in an expression.\n\tFound in %s" % (expr,)
                    )
                self._args.append(e.template)
                _hash.append(id(e.template._set))
            finally:
                logging.disable(logging.NOTSET)

        self._hash = tuple(_hash)

    def nargs(self):
        return len(self._args)

    def arg(self, i):
        return self._args[i]

    @property
    def base(self):
        return self._base

    @property
    def args(self):
        return self._args

    def __hash__(self):
        return hash(self._hash)

    def __eq__(self, other):
        if type(other) is _GetItemIndexer:
            return self._hash == other._hash
        else:
            return False

    def __str__(self):
        return "%s[%s]" % (self._base.name, ','.join(str(x) for x in self._args))


def substitute_getitem_with_param(expr, _map):
    """A simple substituter to replace _GetItem nodes with mutable Params.

    This substituter will replace all GetItemExpression nodes with a
    new Param.  For example, this method will create expressions
    suitable for passing to DAE integrators
    """
    import pyomo.core.base.param

    if type(expr) is IndexTemplate:
        return expr

    _id = _GetItemIndexer(expr)
    if _id not in _map:
        _map[_id] = pyomo.core.base.param.Param(mutable=True)
        _map[_id].construct()
        _map[_id]._name = "%s[%s]" % (_id.base.name, ','.join(str(x) for x in _id.args))
    return _map[_id]


def substitute_template_with_value(expr):
    """A simple substituter to expand expression for current template

    This substituter will replace all GetItemExpression / IndexTemplate
    nodes with the actual _ComponentData based on the current value of
    the IndexTemplate(s)

    """

    if type(expr) is IndexTemplate:
        return as_numeric(expr())
    else:
        return resolve_template(expr)


class _set_iterator_template_generator:
    """Replacement iterator that returns IndexTemplates

    In order to generate template expressions, we hijack the normal Set
    iteration mechanisms so that this iterator is returned instead of
    the usual iterator.  This iterator will return IndexTemplate
    object(s) instead of the actual Set items the first time next() is
    called.
    """

    def __init__(self, _set, context):
        self._set = _set
        self.context = context

    def __iter__(self):
        return self

    def __next__(self):
        # Prevent context from ever being called more than once
        if self.context is None:
            raise StopIteration()
        context, self.context = self.context, None

        _set = self._set
        if _set.is_expression_type():
            d = _reduce_template_to_component(_set).dimen
        else:
            d = _set.dimen
        grp = context.next_group()
        if d is None or type(d) is not int:
            idx = (IndexTemplate(_set, None, context.next_id(), grp),)
        else:
            idx = tuple(
                IndexTemplate(_set, i, context.next_id(), grp) for i in range(d)
            )
        context.cache.append(idx)
        if len(idx) == 1:
            return idx[0]
        else:
            return idx

    next = __next__


class _template_iter_context:
    """Manage the iteration context when generating templatized rules

    This class manages the context tracking when generating templatized
    rules.  It has two methods (`sum_template` and `get_iter`) that
    replace standard functions / methods (`sum` and
    :py:meth:`_FiniteSetMixin.__iter__`, respectively).  It also tracks
    unique identifiers for IndexTemplate objects and their groupings
    within `sum()` generators.
    """

    def __init__(self):
        self.cache = []
        self._id = 0
        self._group = 0

    def get_iter(self, _set):
        return _set_iterator_template_generator(_set, self)

    def npop_cache(self, n):
        result = self.cache[-n:]
        self.cache[-n:] = []
        return result

    def next_id(self):
        self._id += 1
        return self._id

    def next_group(self):
        self._group += 1
        return self._group

    def sum_template(self, generator):
        init_cache = len(self.cache)
        expr = next(generator)
        final_cache = len(self.cache)
        return TemplateSumExpression((expr,), self.npop_cache(final_cache - init_cache))


class _template_iter_manager:
    class _iter_wrapper:
        __slots__ = ('_class', '_iter', '_old_iter')

        def __init__(self, cls, context):
            def _iter_fcn(obj):
                return context.get_iter(obj)

            self._class = cls
            self._old_iter = cls.__iter__
            self._iter = _iter_fcn

        def acquire(self):
            self._class.__iter__ = self._iter

        def release(self):
            self._class.__iter__ = self._old_iter

    class _pause_template_iter_manager:
        __slots__ = ('iter_manager',)

        def __init__(self, iter_manager):
            self.iter_manager = iter_manager

        def __enter__(self):
            self.iter_manager.release()
            return self

        def __exit__(self, et, ev, tb):
            self.iter_manager.acquire()

    def __init__(self):
        self.paused = True
        self.context = None
        self.iters = None
        self.builtin_sum = builtins.sum

    def init(self, context, *iter_fcns):
        assert self.context is None
        self.context = context
        self.iters = [self._iter_wrapper(it, context) for it in iter_fcns]
        return self

    def acquire(self):
        assert self.paused
        self.paused = False
        builtins.sum = self.context.sum_template
        for it in self.iters:
            it.acquire()

    def release(self):
        assert not self.paused
        self.paused = True
        builtins.sum = self.builtin_sum
        for it in self.iters:
            it.release()

    def __enter__(self):
        assert self.context
        self.acquire()
        return self

    def __exit__(self, et, ev, tb):
        self.release()
        self.context = None
        self.iters = None

    def pause(self):
        if self.paused:
            return nullcontext()
        else:
            return self._pause_template_iter_manager(self)


# Global manager for coordinating overriding set iteration
_TemplateIterManager = _template_iter_manager()


def templatize_rule(block, rule, index_set):
    import pyomo.core.base.set

    context = _template_iter_context()
    internal_error = None
    try:
        # Override Set iteration to return IndexTemplates
        with _TemplateIterManager.init(
            context,
            pyomo.core.base.set._FiniteSetMixin,
            GetItemExpression,
            GetAttrExpression,
        ):
            # Get the index templates needed for calling the rule
            if index_set is not None:
                # Note, do not rely on the __iter__ overload, as non-finite
                # Sets don't have an __iter__.
                indices = next(iter(context.get_iter(index_set)))
                try:
                    context.cache.pop()
                except IndexError:
                    assert indices is None
                    indices = ()
            else:
                indices = ()
            if type(indices) is not tuple:
                indices = (indices,)
            # Call the rule, returning the template expression and the
            # top-level IndexTemplate(s) generated when calling the rule.
            #
            # TBD: Should this just return a "FORALL()" expression node that
            # behaves similarly to the GetItemExpression node?
            return rule(block, indices), indices
    except:
        internal_error = sys.exc_info()
        raise
    finally:
        if len(context.cache):
            if internal_error is not None:
                logger.error(
                    "The following exception was raised when "
                    "templatizing the rule '%s':\n\t%s" % (rule.name, internal_error[1])
                )
            raise TemplateExpressionError(
                None,
                "Explicit iteration (for loops) over Sets is not supported "
                "by template expressions.  Encountered loop over %s"
                % (context.cache[-1][0]._set,),
            )
    return None, indices


def templatize_constraint(con):
    expr, indices = templatize_rule(con.parent_block(), con.rule, con.index_set())
    if expr.__class__ is tuple:
        expr = tuple_to_relational_expr(expr)
    return expr, indices



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.common.tempfiles import TempfileManager

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)
from pyomo.common.dependencies.scipy import sparse as spa

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest("Pynumero needs the ASL extension to run cyipopt tests")

from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import cyipopt_available

from pyomo.contrib.pynumero.interfaces.external_grey_box import (
    ExternalGreyBoxModel,
    ExternalGreyBoxBlock,
)
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoGreyBoxNLP
from pyomo.contrib.pynumero.interfaces.tests.compare_utils import (
    check_vectors_specific_order,
    check_sparse_matrix_specific_order,
)
import pyomo.contrib.pynumero.interfaces.tests.external_grey_box_models as ex_models


class TestExternalGreyBoxModel(unittest.TestCase):
    def test_pressure_drop_single_output(self):
        egbm = ex_models.PressureDropSingleOutput()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, [])
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([], dtype=np.float64))
        with self.assertRaises(AssertionError):
            egbm.set_equality_constraint_multipliers(np.asarray([1], dtype=np.float64))
        egbm.set_output_constraint_multipliers(np.asarray([5], dtype=np.float64))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_equality_constraints()

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([28], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_equality_constraints()

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(
            np.array_equal(jac_o.row, np.asarray([0, 0, 0], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.col, np.asarray([0, 1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.data, np.asarray([1, -36, -48], dtype=np.float64))
        )

        with self.assertRaises(AttributeError):
            eq_hess = egbm.evaluate_hessian_equality_constraints()

        with self.assertRaises(AttributeError):
            outputs_hess = egbm.evaluate_hessian_outputs()

    def test_pressure_drop_single_output_with_hessian(self):
        egbm = ex_models.PressureDropSingleOutputWithHessian()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, [])
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([], dtype=np.float64))
        with self.assertRaises(AssertionError):
            egbm.set_equality_constraint_multipliers(np.asarray([1], dtype=np.float64))
        egbm.set_output_constraint_multipliers(np.asarray([5], dtype=np.float64))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_equality_constraints()

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([28], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_equality_constraints()

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(
            np.array_equal(jac_o.row, np.asarray([0, 0, 0], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.col, np.asarray([0, 1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.data, np.asarray([1, -36, -48], dtype=np.float64))
        )

        with self.assertRaises(AttributeError):
            eq_hess = egbm.evaluate_hessian_equality_constraints()
        outputs_hess = egbm.evaluate_hessian_outputs()
        self.assertTrue(
            np.array_equal(outputs_hess.row, np.asarray([2, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(outputs_hess.col, np.asarray([1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(
                outputs_hess.data,
                np.asarray([5 * (-8 * 3), 5 * (-8 * 2)], dtype=np.int64),
            )
        )

    def test_pressure_drop_single_equality(self):
        egbm = ex_models.PressureDropSingleEquality()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'Pout'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop'])
        output_names = egbm.output_names()
        self.assertEqual(output_names, [])

        egbm.set_input_values(np.asarray([100, 2, 3, 50], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([5], dtype=np.float64))
        with self.assertRaises(AssertionError):
            egbm.set_output_constraint_multipliers(np.asarray([1], dtype=np.float64))

        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([22], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_outputs()

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_outputs()

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(
            np.array_equal(jac_eq.row, np.asarray([0, 0, 0, 0], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_eq.col, np.asarray([0, 1, 2, 3], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_eq.data, np.asarray([-1, 36, 48, 1], dtype=np.float64))
        )

        with self.assertRaises(AttributeError):
            eq_hess = egbm.evaluate_hessian_equality_constraints()
        with self.assertRaises(AttributeError):
            outputs_hess = egbm.evaluate_hessian_outputs()

    def test_pressure_drop_single_equality_with_hessian(self):
        egbm = ex_models.PressureDropSingleEqualityWithHessian()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'Pout'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop'])
        output_names = egbm.output_names()
        self.assertEqual(output_names, [])

        egbm.set_input_values(np.asarray([100, 2, 3, 50], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([5], dtype=np.float64))
        with self.assertRaises(AssertionError):
            egbm.set_output_constraint_multipliers(np.asarray([1], dtype=np.float64))

        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([22], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_outputs()

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_outputs()

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(
            np.array_equal(jac_eq.row, np.asarray([0, 0, 0, 0], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_eq.col, np.asarray([0, 1, 2, 3], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_eq.data, np.asarray([-1, 36, 48, 1], dtype=np.float64))
        )

        eq_hess = egbm.evaluate_hessian_equality_constraints()
        with self.assertRaises(AttributeError):
            outputs_hess = egbm.evaluate_hessian_outputs()
        self.assertTrue(np.array_equal(eq_hess.row, np.asarray([2, 2], dtype=np.int64)))
        self.assertTrue(np.array_equal(eq_hess.col, np.asarray([1, 2], dtype=np.int64)))
        self.assertTrue(
            np.array_equal(
                eq_hess.data, np.asarray([5 * (8 * 3), 5 * (8 * 2)], dtype=np.float64)
            )
        )

    def test_pressure_drop_two_outputs(self):
        egbm = ex_models.PressureDropTwoOutputs()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual([], eq_con_names)
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['P2', 'Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([], dtype=np.float64))
        # this one should fail
        with self.assertRaises(AssertionError):
            egbm.set_equality_constraint_multipliers(np.asarray([1], dtype=np.float64))

        egbm.set_output_constraint_multipliers(np.asarray([3.0, 5.0], dtype=np.float64))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_equality_constraints()

        #   u = [Pin, c, F]
        #   o = [P2, Pout]
        #   h_eq(u) = {empty}
        #   h_o(u) = [Pin - 2*c*F^2]
        #            [Pin - 4*c*F^2]

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([64, 28], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_equality_constraints()

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(
            np.array_equal(jac_o.row, np.asarray([0, 0, 0, 1, 1, 1], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.col, np.asarray([0, 1, 2, 0, 1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(
                jac_o.data, np.asarray([1, -18, -24, 1, -36, -48], dtype=np.float64)
            )
        )

        with self.assertRaises(AttributeError):
            hess_eq = egbm.evaluate_hessian_equality_constraints()
        with self.assertRaises(AttributeError):
            hess_outputs = egbm.evaluate_hessian_outputs()

    def test_pressure_drop_two_outputs_with_hessian(self):
        egbm = ex_models.PressureDropTwoOutputsWithHessian()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual([], eq_con_names)
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['P2', 'Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([], dtype=np.float64))
        # this one should fail
        with self.assertRaises(AssertionError):
            egbm.set_equality_constraint_multipliers(np.asarray([1], dtype=np.float64))

        egbm.set_output_constraint_multipliers(np.asarray([3.0, 5.0], dtype=np.float64))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_equality_constraints()

        #   u = [Pin, c, F]
        #   o = [P2, Pout]
        #   h_eq(u) = {empty}
        #   h_o(u) = [Pin - 2*c*F^2]
        #            [Pin - 4*c*F^2]

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([64, 28], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_equality_constraints()

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(
            np.array_equal(jac_o.row, np.asarray([0, 0, 0, 1, 1, 1], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.col, np.asarray([0, 1, 2, 0, 1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(
                jac_o.data, np.asarray([1, -18, -24, 1, -36, -48], dtype=np.float64)
            )
        )

        with self.assertRaises(AttributeError):
            hess_eq = egbm.evaluate_hessian_equality_constraints()
        hess = egbm.evaluate_hessian_outputs()
        self.assertTrue(np.array_equal(hess.row, np.asarray([2, 2], dtype=np.int64)))
        self.assertTrue(np.array_equal(hess.col, np.asarray([1, 2], dtype=np.int64)))
        self.assertTrue(
            np.array_equal(hess.data, np.asarray([-156.0, -104.0], dtype=np.float64))
        )

    def test_pressure_drop_two_equalities(self):
        egbm = ex_models.PressureDropTwoEqualities()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'P2', 'Pout'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop2', 'pdropout'])
        output_names = egbm.output_names()
        self.assertEqual([], output_names)

        egbm.set_input_values(np.asarray([100, 2, 3, 20, 50], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([3, 5], dtype=np.float64))
        egbm.set_output_constraint_multipliers(np.asarray([]))
        # this one should fail
        with self.assertRaises(AssertionError):
            egbm.set_output_constraint_multipliers(np.asarray([1], dtype=np.float64))

        #   u = [Pin, c, F, P2, Pout]
        #   o = {empty}
        #   h_eq(u) = [P2 - (Pin - 2*c*F^2]
        #             [Pout - (P2 - 2*c*F^2]
        #   h_o(u) = {empty}
        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([-44, 66], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_outputs()

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_outputs()

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(
            np.array_equal(
                jac_eq.row, np.asarray([0, 0, 0, 0, 1, 1, 1, 1], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.col, np.asarray([0, 1, 2, 3, 1, 2, 3, 4], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.data,
                np.asarray([-1, 18, 24, 1, 18, 24, -1, 1], dtype=np.float64),
            )
        )

        with self.assertRaises(AttributeError):
            hess_outputs = egbm.evaluate_hessian_outputs()
        with self.assertRaises(AttributeError):
            hess = egbm.evaluate_hessian_equality_constraints()

    def test_pressure_drop_two_equalities_with_hessian(self):
        egbm = ex_models.PressureDropTwoEqualitiesWithHessian()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'P2', 'Pout'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop2', 'pdropout'])
        output_names = egbm.output_names()
        self.assertEqual([], output_names)

        egbm.set_input_values(np.asarray([100, 2, 3, 20, 50], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([3, 5], dtype=np.float64))
        egbm.set_output_constraint_multipliers(np.asarray([]))
        # this one should fail
        with self.assertRaises(AssertionError):
            egbm.set_output_constraint_multipliers(np.asarray([1], dtype=np.float64))

        #   u = [Pin, c, F, P2, Pout]
        #   o = {empty}
        #   h_eq(u) = [P2 - (Pin - 2*c*F^2]
        #             [Pout - (P2 - 2*c*F^2]
        #   h_o(u) = {empty}
        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([-44, 66], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_outputs()

        with self.assertRaises(NotImplementedError):
            tmp = egbm.evaluate_jacobian_outputs()

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(
            np.array_equal(
                jac_eq.row, np.asarray([0, 0, 0, 0, 1, 1, 1, 1], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.col, np.asarray([0, 1, 2, 3, 1, 2, 3, 4], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.data,
                np.asarray([-1, 18, 24, 1, 18, 24, -1, 1], dtype=np.float64),
            )
        )

        with self.assertRaises(AttributeError):
            hess_outputs = egbm.evaluate_hessian_outputs()
        hess = egbm.evaluate_hessian_equality_constraints()
        self.assertTrue(np.array_equal(hess.row, np.asarray([2, 2], dtype=np.int64)))
        self.assertTrue(np.array_equal(hess.col, np.asarray([1, 2], dtype=np.int64)))
        self.assertTrue(
            np.array_equal(hess.data, np.asarray([96.0, 64.0], dtype=np.float64))
        )

    def test_pressure_drop_two_equalities_two_outputs(self):
        #   u = [Pin, c, F, P1, P3]
        #   o = {P2, Pout}
        #   h_eq(u) = [P1 - (Pin - c*F^2]
        #             [P3 - (Pin - 2*c*F^2]
        #   h_o(u) = [P1 - c*F^2]
        #            [Pin - 4*c*F^2]
        egbm = ex_models.PressureDropTwoEqualitiesTwoOutputs()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'P1', 'P3'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop1', 'pdrop3'])
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['P2', 'Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3, 80, 70], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([2, 4], dtype=np.float64))
        egbm.set_output_constraint_multipliers(np.asarray([7, 9], dtype=np.float64))
        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([-2, 26], dtype=np.float64)))

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([62, 28], dtype=np.float64)))

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(
            np.array_equal(
                jac_eq.row, np.asarray([0, 0, 0, 0, 1, 1, 1, 1], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.col, np.asarray([0, 1, 2, 3, 1, 2, 3, 4], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.data, np.asarray([-1, 9, 12, 1, 18, 24, -1, 1], dtype=np.float64)
            )
        )

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(
            np.array_equal(jac_o.row, np.asarray([0, 0, 0, 1, 1, 1], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.col, np.asarray([1, 2, 3, 0, 1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(
                jac_o.data, np.asarray([-9, -12, 1, 1, -36, -48], dtype=np.float64)
            )
        )

        with self.assertRaises(AttributeError):
            hess = egbm.evaluate_hessian_equality_constraints()
        with self.assertRaises(AttributeError):
            hess = egbm.evaluate_hessian_outputs()

    def test_pressure_drop_two_equalities_two_outputs_with_hessian(self):
        #   u = [Pin, c, F, P1, P3]
        #   o = {P2, Pout}
        #   h_eq(u) = [P1 - (Pin - c*F^2]
        #             [P3 - (Pin - 2*c*F^2]
        #   h_o(u) = [P1 - c*F^2]
        #            [Pin - 4*c*F^2]
        egbm = ex_models.PressureDropTwoEqualitiesTwoOutputsWithHessian()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'P1', 'P3'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop1', 'pdrop3'])
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['P2', 'Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3, 80, 70], dtype=np.float64))
        egbm.set_equality_constraint_multipliers(np.asarray([2, 4], dtype=np.float64))
        egbm.set_output_constraint_multipliers(np.asarray([7, 9], dtype=np.float64))
        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([-2, 26], dtype=np.float64)))

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([62, 28], dtype=np.float64)))

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(
            np.array_equal(
                jac_eq.row, np.asarray([0, 0, 0, 0, 1, 1, 1, 1], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.col, np.asarray([0, 1, 2, 3, 1, 2, 3, 4], dtype=np.int64)
            )
        )
        self.assertTrue(
            np.array_equal(
                jac_eq.data, np.asarray([-1, 9, 12, 1, 18, 24, -1, 1], dtype=np.float64)
            )
        )

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(
            np.array_equal(jac_o.row, np.asarray([0, 0, 0, 1, 1, 1], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(jac_o.col, np.asarray([1, 2, 3, 0, 1, 2], dtype=np.int64))
        )
        self.assertTrue(
            np.array_equal(
                jac_o.data, np.asarray([-9, -12, 1, 1, -36, -48], dtype=np.float64)
            )
        )

        hess = egbm.evaluate_hessian_equality_constraints()
        self.assertTrue(np.array_equal(hess.row, np.asarray([2, 2], dtype=np.int64)))
        self.assertTrue(np.array_equal(hess.col, np.asarray([1, 2], dtype=np.int64)))
        self.assertTrue(
            np.array_equal(hess.data, np.asarray([60.0, 40.0], dtype=np.float64))
        )

        hess = egbm.evaluate_hessian_outputs()
        self.assertTrue(np.array_equal(hess.row, np.asarray([2, 2], dtype=np.int64)))
        self.assertTrue(np.array_equal(hess.col, np.asarray([1, 2], dtype=np.int64)))
        self.assertTrue(
            np.array_equal(hess.data, np.asarray([-258, -172], dtype=np.float64))
        )


"""
    def test_pressure_drop_two_equalities_two_outputs_no_hessian(self):
        #   u = [Pin, c, F, P1, P3]
        #   o = {P2, Pout}
        #   h_eq(u) = [P1 - (Pin - c*F^2]
        #             [P3 - (Pin - 2*c*F^2]
        #   h_o(u) = [P1 - c*F^2]
        #            [Pin - 4*c*F^2]
        egbm = ex_models.PressureDropTwoEqualitiesTwoOutputsNoHessian()
        input_names = egbm.input_names()
        self.assertEqual(input_names, ['Pin', 'c', 'F', 'P1', 'P3'])
        eq_con_names = egbm.equality_constraint_names()
        self.assertEqual(eq_con_names, ['pdrop1', 'pdrop3'])
        output_names = egbm.output_names()
        self.assertEqual(output_names, ['P2', 'Pout'])

        egbm.set_input_values(np.asarray([100, 2, 3, 80, 70], dtype=np.float64))
        with self.assertRaises(NotImplementedError):
            egbm.set_equality_constraint_multipliers(np.asarray([2, 4], dtype=np.float64))
        with self.assertRaises(NotImplementedError):
            egbm.set_output_constraint_multipliers(np.asarray([7, 9], dtype=np.float64))
            
        eq = egbm.evaluate_equality_constraints()
        self.assertTrue(np.array_equal(eq, np.asarray([-2, 26], dtype=np.float64)))

        o = egbm.evaluate_outputs()
        self.assertTrue(np.array_equal(o, np.asarray([62, 28], dtype=np.float64)))

        jac_eq = egbm.evaluate_jacobian_equality_constraints()
        self.assertTrue(np.array_equal(jac_eq.row, np.asarray([0,0,0,0,1,1,1,1], dtype=np.int64)))
        self.assertTrue(np.array_equal(jac_eq.col, np.asarray([0,1,2,3,1,2,3,4], dtype=np.int64)))
        self.assertTrue(np.array_equal(jac_eq.data, np.asarray([-1, 9, 12, 1, 18, 24, -1, 1], dtype=np.float64)))

        jac_o = egbm.evaluate_jacobian_outputs()
        self.assertTrue(np.array_equal(jac_o.row, np.asarray([0,0,0,1,1,1], dtype=np.int64)))
        self.assertTrue(np.array_equal(jac_o.col, np.asarray([1,2,3,0,1,2], dtype=np.int64)))
        self.assertTrue(np.array_equal(jac_o.data, np.asarray([-9, -12, 1, 1, -36, -48], dtype=np.float64)))

        with self.assertRaises(NotImplementedError):
            hess = egbm.evaluate_hessian_equality_constraints()

        with self.assertRaises(NotImplementedError):
            hess = egbm.evaluate_hessian_outputs()
"""


# TODO: make this work even if there is only external and no variables anywhere in pyomo part
class TestPyomoGreyBoxNLP(unittest.TestCase):
    def test_error_no_variables(self):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_models.PressureDropSingleOutput())
        m.obj = pyo.Objective(expr=1)
        with self.assertRaises(ValueError):
            pyomo_nlp = PyomoGreyBoxNLP(m)

    def test_error_fixed_inputs_outputs(self):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_models.PressureDropSingleOutput())
        m.egb.inputs['Pin'].fix(100)
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        with self.assertRaises(NotImplementedError):
            pyomo_nlp = PyomoGreyBoxNLP(m)

        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_models.PressureDropTwoOutputs())
        m.egb.outputs['P2'].fix(50)
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        with self.assertRaises(NotImplementedError):
            pyomo_nlp = PyomoGreyBoxNLP(m)

    def test_pressure_drop_single_output(self):
        self._test_pressure_drop_single_output(
            ex_models.PressureDropSingleOutput(), False
        )
        self._test_pressure_drop_single_output(
            ex_models.PressureDropSingleOutputWithHessian(), True
        )

    def _test_pressure_drop_single_output(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(0)
        m.egb.outputs['Pout'].setub(100)
        # m.dummy = pyo.Constraint(expr=sum(m.egb.inputs[i] for i in m.egb.inputs) + sum(m.egb.outputs[i] for i in m.egb.outputs) <= 1e6)
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)

        pyomo_nlp = PyomoGreyBoxNLP(m)

        self.assertEqual(4, pyomo_nlp.n_primals())
        self.assertEqual(1, pyomo_nlp.n_constraints())
        self.assertEqual(4, pyomo_nlp.nnz_jacobian())
        if hessian_support:
            self.assertEqual(3, pyomo_nlp.nnz_hessian_lag())

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.outputs[Pout]',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = ['egb.Pout_con']
        c_order = pyomo_nlp.constraint_names()

        xlb = pyomo_nlp.primals_lb()
        comparison_xlb = np.asarray([50, 1, 1, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, xlb, x_order, comparison_xlb, comparison_x_order
        )
        xub = pyomo_nlp.primals_ub()
        comparison_xub = np.asarray([150, 5, 5, 100], dtype=np.float64)
        check_vectors_specific_order(
            self, xub, x_order, comparison_xub, comparison_x_order
        )
        clb = pyomo_nlp.constraints_lb()
        comparison_clb = np.asarray([0], dtype=np.float64)
        check_vectors_specific_order(
            self, clb, c_order, comparison_clb, comparison_c_order
        )
        cub = pyomo_nlp.constraints_ub()
        comparison_cub = np.asarray([0], dtype=np.float64)
        check_vectors_specific_order(
            self, cub, c_order, comparison_cub, comparison_c_order
        )

        xinit = pyomo_nlp.init_primals()
        comparison_xinit = np.asarray([100, 2, 3, 50], dtype=np.float64)
        check_vectors_specific_order(
            self, xinit, x_order, comparison_xinit, comparison_x_order
        )
        duals_init = pyomo_nlp.init_duals()
        comparison_duals_init = np.asarray([0], dtype=np.float64)
        check_vectors_specific_order(
            self, duals_init, c_order, comparison_duals_init, comparison_c_order
        )

        self.assertEqual(4, len(pyomo_nlp.create_new_vector('primals')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('constraints')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('duals')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('eq_constraints')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('ineq_constraints')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('duals_eq')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('duals_ineq')))

        pyomo_nlp.set_primals(np.asarray([1, 2, 3, 4], dtype=np.float64))
        x = pyomo_nlp.get_primals()
        self.assertTrue(np.array_equal(x, np.asarray([1, 2, 3, 4], dtype=np.float64)))
        pyomo_nlp.set_primals(pyomo_nlp.init_primals())

        pyomo_nlp.set_duals(np.asarray([42], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([42], dtype=np.float64)))
        pyomo_nlp.set_duals(np.asarray([21], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([21], dtype=np.float64)))

        fac = pyomo_nlp.get_obj_factor()
        self.assertEqual(fac, 1)
        pyomo_nlp.set_obj_factor(42)
        self.assertEqual(pyomo_nlp.get_obj_factor(), 42)
        pyomo_nlp.set_obj_factor(1)

        f = pyomo_nlp.evaluate_objective()
        self.assertEqual(f, 900)

        gradf = pyomo_nlp.evaluate_grad_objective()
        comparison_gradf = np.asarray([0, 0, 0, 60], dtype=np.float64)
        check_vectors_specific_order(
            self, gradf, x_order, comparison_gradf, comparison_x_order
        )
        c = pyomo_nlp.evaluate_constraints()
        comparison_c = np.asarray([-22], dtype=np.float64)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)
        c = np.zeros(1)
        pyomo_nlp.evaluate_constraints(out=c)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)

        j = pyomo_nlp.evaluate_jacobian()
        comparison_j = np.asarray([[1, -36, -48, -1]])
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        j = 2.0 * j
        pyomo_nlp.evaluate_jacobian(out=j)
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        if hessian_support:
            h = pyomo_nlp.evaluate_hessian_lag()
            self.assertTrue(h.shape == (4, 4))
            comparison_h = np.asarray(
                [
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, -8 * 3 * 21, -8 * 2 * 21, 0],
                    [0, 0, 0, 2 * 1],
                ],
                dtype=np.float64,
            )
            check_sparse_matrix_specific_order(
                self,
                h,
                x_order,
                x_order,
                comparison_h,
                comparison_x_order,
                comparison_x_order,
            )
        else:
            with self.assertRaises(AttributeError):
                h = pyomo_nlp.evaluate_hessian_lag()

    def test_pressure_drop_single_equality(self):
        self._test_pressure_drop_single_equality(
            ex_models.PressureDropSingleEquality(), False
        )
        self._test_pressure_drop_single_equality(
            ex_models.PressureDropSingleEqualityWithHessian(), True
        )

    def _test_pressure_drop_single_equality(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['Pout'].value = 50
        m.egb.inputs['Pout'].setlb(0)
        m.egb.inputs['Pout'].setub(100)
        m.obj = pyo.Objective(expr=(m.egb.inputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)

        self.assertEqual(4, pyomo_nlp.n_primals())
        self.assertEqual(1, pyomo_nlp.n_constraints())
        self.assertEqual(4, pyomo_nlp.nnz_jacobian())
        if hessian_support:
            self.assertEqual(3, pyomo_nlp.nnz_hessian_lag())

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[Pout]',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = ['egb.pdrop']
        c_order = pyomo_nlp.constraint_names()

        xlb = pyomo_nlp.primals_lb()
        comparison_xlb = np.asarray([50, 1, 1, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, xlb, x_order, comparison_xlb, comparison_x_order
        )
        xub = pyomo_nlp.primals_ub()
        comparison_xub = np.asarray([150, 5, 5, 100], dtype=np.float64)
        check_vectors_specific_order(
            self, xub, x_order, comparison_xub, comparison_x_order
        )
        clb = pyomo_nlp.constraints_lb()
        comparison_clb = np.asarray([0], dtype=np.float64)
        check_vectors_specific_order(
            self, clb, c_order, comparison_clb, comparison_c_order
        )
        cub = pyomo_nlp.constraints_ub()
        comparison_cub = np.asarray([0], dtype=np.float64)
        check_vectors_specific_order(
            self, cub, c_order, comparison_cub, comparison_c_order
        )

        xinit = pyomo_nlp.init_primals()
        comparison_xinit = np.asarray([100, 2, 3, 50], dtype=np.float64)
        check_vectors_specific_order(
            self, xinit, x_order, comparison_xinit, comparison_x_order
        )
        duals_init = pyomo_nlp.init_duals()
        comparison_duals_init = np.asarray([0], dtype=np.float64)
        check_vectors_specific_order(
            self, duals_init, c_order, comparison_duals_init, comparison_c_order
        )

        self.assertEqual(4, len(pyomo_nlp.create_new_vector('primals')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('constraints')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('duals')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('eq_constraints')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('ineq_constraints')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('duals_eq')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('duals_ineq')))

        pyomo_nlp.set_primals(np.asarray([1, 2, 3, 4], dtype=np.float64))
        x = pyomo_nlp.get_primals()
        self.assertTrue(np.array_equal(x, np.asarray([1, 2, 3, 4], dtype=np.float64)))
        pyomo_nlp.set_primals(pyomo_nlp.init_primals())

        pyomo_nlp.set_duals(np.asarray([42], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([42], dtype=np.float64)))
        pyomo_nlp.set_duals(np.asarray([21], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([21], dtype=np.float64)))

        fac = pyomo_nlp.get_obj_factor()
        self.assertEqual(fac, 1)
        pyomo_nlp.set_obj_factor(42)
        self.assertEqual(pyomo_nlp.get_obj_factor(), 42)
        pyomo_nlp.set_obj_factor(1)

        f = pyomo_nlp.evaluate_objective()
        self.assertEqual(f, 900)

        gradf = pyomo_nlp.evaluate_grad_objective()
        comparison_gradf = np.asarray([0, 0, 0, 60], dtype=np.float64)
        check_vectors_specific_order(
            self, gradf, x_order, comparison_gradf, comparison_x_order
        )
        c = pyomo_nlp.evaluate_constraints()
        comparison_c = np.asarray([22], dtype=np.float64)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)
        c = np.zeros(1)
        pyomo_nlp.evaluate_constraints(out=c)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)

        j = pyomo_nlp.evaluate_jacobian()
        comparison_j = np.asarray([[-1, 36, 48, 1]])
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        j = 2.0 * j
        pyomo_nlp.evaluate_jacobian(out=j)
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        if hessian_support:
            h = pyomo_nlp.evaluate_hessian_lag()
            self.assertTrue(h.shape == (4, 4))
            comparison_h = np.asarray(
                [
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 8 * 3 * 21, 8 * 2 * 21, 0],
                    [0, 0, 0, 2 * 1],
                ],
                dtype=np.float64,
            )
            check_sparse_matrix_specific_order(
                self,
                h,
                x_order,
                x_order,
                comparison_h,
                comparison_x_order,
                comparison_x_order,
            )
        else:
            with self.assertRaises(AttributeError):
                h = pyomo_nlp.evaluate_hessian_lag()

    def test_pressure_drop_two_outputs(self):
        self._test_pressure_drop_two_outputs(ex_models.PressureDropTwoOutputs(), False)
        self._test_pressure_drop_two_outputs(
            ex_models.PressureDropTwoOutputsWithHessian(), True
        )

    def _test_pressure_drop_two_outputs(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.outputs['P2'].value = 80
        m.egb.outputs['P2'].setlb(10)
        m.egb.outputs['P2'].setub(90)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(0)
        m.egb.outputs['Pout'].setub(100)
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)

        self.assertEqual(5, pyomo_nlp.n_primals())
        self.assertEqual(2, pyomo_nlp.n_constraints())
        self.assertEqual(8, pyomo_nlp.nnz_jacobian())
        if hessian_support:
            self.assertEqual(3, pyomo_nlp.nnz_hessian_lag())

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.outputs[P2]',
            'egb.outputs[Pout]',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = ['egb.P2_con', 'egb.Pout_con']
        c_order = pyomo_nlp.constraint_names()

        xlb = pyomo_nlp.primals_lb()
        comparison_xlb = np.asarray([50, 1, 1, 10, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, xlb, x_order, comparison_xlb, comparison_x_order
        )
        xub = pyomo_nlp.primals_ub()
        comparison_xub = np.asarray([150, 5, 5, 90, 100], dtype=np.float64)
        check_vectors_specific_order(
            self, xub, x_order, comparison_xub, comparison_x_order
        )
        clb = pyomo_nlp.constraints_lb()
        comparison_clb = np.asarray([0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, clb, c_order, comparison_clb, comparison_c_order
        )
        cub = pyomo_nlp.constraints_ub()
        comparison_cub = np.asarray([0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, cub, c_order, comparison_cub, comparison_c_order
        )

        xinit = pyomo_nlp.init_primals()
        comparison_xinit = np.asarray([100, 2, 3, 80, 50], dtype=np.float64)
        check_vectors_specific_order(
            self, xinit, x_order, comparison_xinit, comparison_x_order
        )
        duals_init = pyomo_nlp.init_duals()
        comparison_duals_init = np.asarray([0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, duals_init, c_order, comparison_duals_init, comparison_c_order
        )

        self.assertEqual(5, len(pyomo_nlp.create_new_vector('primals')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('constraints')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('duals')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('eq_constraints')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('ineq_constraints')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('duals_eq')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('duals_ineq')))

        pyomo_nlp.set_primals(np.asarray([1, 2, 3, 4, 5], dtype=np.float64))
        x = pyomo_nlp.get_primals()
        self.assertTrue(
            np.array_equal(x, np.asarray([1, 2, 3, 4, 5], dtype=np.float64))
        )
        pyomo_nlp.set_primals(pyomo_nlp.init_primals())

        pyomo_nlp.set_duals(np.asarray([42, 10], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([42, 10], dtype=np.float64)))
        pyomo_nlp.set_duals(np.asarray([21, 5], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([21, 5], dtype=np.float64)))

        fac = pyomo_nlp.get_obj_factor()
        self.assertEqual(fac, 1)
        pyomo_nlp.set_obj_factor(42)
        self.assertEqual(pyomo_nlp.get_obj_factor(), 42)
        pyomo_nlp.set_obj_factor(1)

        f = pyomo_nlp.evaluate_objective()
        self.assertEqual(f, 900)

        gradf = pyomo_nlp.evaluate_grad_objective()
        comparison_gradf = np.asarray([0, 0, 0, 0, 60], dtype=np.float64)
        check_vectors_specific_order(
            self, gradf, x_order, comparison_gradf, comparison_x_order
        )
        c = pyomo_nlp.evaluate_constraints()
        comparison_c = np.asarray([-16, -22], dtype=np.float64)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)
        c = np.zeros(2)
        pyomo_nlp.evaluate_constraints(out=c)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)

        j = pyomo_nlp.evaluate_jacobian()
        comparison_j = np.asarray([[1, -18, -24, -1, 0], [1, -36, -48, 0, -1]])
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        j = 2.0 * j
        pyomo_nlp.evaluate_jacobian(out=j)
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        if hessian_support:
            h = pyomo_nlp.evaluate_hessian_lag()
            self.assertTrue(h.shape == (5, 5))
            comparison_h = np.asarray(
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [
                        0,
                        (-4 * 3 * 21) + (-8 * 3 * 5),
                        (-4 * 2 * 21) + (-8 * 2 * 5),
                        0,
                        0,
                    ],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 2 * 1],
                ],
                dtype=np.float64,
            )
            check_sparse_matrix_specific_order(
                self,
                h,
                x_order,
                x_order,
                comparison_h,
                comparison_x_order,
                comparison_x_order,
            )
        else:
            with self.assertRaises(AttributeError):
                h = pyomo_nlp.evaluate_hessian_lag()

    def test_pressure_drop_two_equalities(self):
        self._test_pressure_drop_two_equalities(
            ex_models.PressureDropTwoEqualities(), False
        )
        self._test_pressure_drop_two_equalities(
            ex_models.PressureDropTwoEqualitiesWithHessian(), True
        )

    def _test_pressure_drop_two_equalities(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['P2'].value = 80
        m.egb.inputs['P2'].setlb(10)
        m.egb.inputs['P2'].setub(90)
        m.egb.inputs['Pout'].value = 50
        m.egb.inputs['Pout'].setlb(0)
        m.egb.inputs['Pout'].setub(100)
        m.obj = pyo.Objective(expr=(m.egb.inputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)

        self.assertEqual(5, pyomo_nlp.n_primals())
        self.assertEqual(2, pyomo_nlp.n_constraints())
        self.assertEqual(8, pyomo_nlp.nnz_jacobian())
        if hessian_support:
            self.assertEqual(3, pyomo_nlp.nnz_hessian_lag())

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[P2]',
            'egb.inputs[Pout]',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = ['egb.pdrop2', 'egb.pdropout']
        c_order = pyomo_nlp.constraint_names()

        xlb = pyomo_nlp.primals_lb()
        comparison_xlb = np.asarray([50, 1, 1, 10, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, xlb, x_order, comparison_xlb, comparison_x_order
        )
        xub = pyomo_nlp.primals_ub()
        comparison_xub = np.asarray([150, 5, 5, 90, 100], dtype=np.float64)
        check_vectors_specific_order(
            self, xub, x_order, comparison_xub, comparison_x_order
        )
        clb = pyomo_nlp.constraints_lb()
        comparison_clb = np.asarray([0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, clb, c_order, comparison_clb, comparison_c_order
        )
        cub = pyomo_nlp.constraints_ub()
        comparison_cub = np.asarray([0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, cub, c_order, comparison_cub, comparison_c_order
        )

        xinit = pyomo_nlp.init_primals()
        comparison_xinit = np.asarray([100, 2, 3, 80, 50], dtype=np.float64)
        check_vectors_specific_order(
            self, xinit, x_order, comparison_xinit, comparison_x_order
        )
        duals_init = pyomo_nlp.init_duals()
        comparison_duals_init = np.asarray([0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, duals_init, c_order, comparison_duals_init, comparison_c_order
        )

        self.assertEqual(5, len(pyomo_nlp.create_new_vector('primals')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('constraints')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('duals')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('eq_constraints')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('ineq_constraints')))
        self.assertEqual(2, len(pyomo_nlp.create_new_vector('duals_eq')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('duals_ineq')))

        pyomo_nlp.set_primals(np.asarray([1, 2, 3, 4, 5], dtype=np.float64))
        x = pyomo_nlp.get_primals()
        self.assertTrue(
            np.array_equal(x, np.asarray([1, 2, 3, 4, 5], dtype=np.float64))
        )
        pyomo_nlp.set_primals(pyomo_nlp.init_primals())

        pyomo_nlp.set_duals(np.asarray([42, 10], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([42, 10], dtype=np.float64)))
        pyomo_nlp.set_duals(np.asarray([21, 5], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([21, 5], dtype=np.float64)))

        fac = pyomo_nlp.get_obj_factor()
        self.assertEqual(fac, 1)
        pyomo_nlp.set_obj_factor(42)
        self.assertEqual(pyomo_nlp.get_obj_factor(), 42)
        pyomo_nlp.set_obj_factor(1)

        f = pyomo_nlp.evaluate_objective()
        self.assertEqual(f, 900)

        gradf = pyomo_nlp.evaluate_grad_objective()
        comparison_gradf = np.asarray([0, 0, 0, 0, 60], dtype=np.float64)
        check_vectors_specific_order(
            self, gradf, x_order, comparison_gradf, comparison_x_order
        )
        c = pyomo_nlp.evaluate_constraints()
        comparison_c = np.asarray([16, 6], dtype=np.float64)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)
        c = np.zeros(2)
        pyomo_nlp.evaluate_constraints(out=c)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)

        j = pyomo_nlp.evaluate_jacobian()
        comparison_j = np.asarray([[-1, 18, 24, 1, 0], [0, 18, 24, -1, 1]])
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        j = 2.0 * j
        pyomo_nlp.evaluate_jacobian(out=j)
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        if hessian_support:
            h = pyomo_nlp.evaluate_hessian_lag()
            self.assertTrue(h.shape == (5, 5))
            comparison_h = np.asarray(
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, (4 * 3 * 21) + (4 * 3 * 5), (4 * 2 * 21) + (4 * 2 * 5), 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 2 * 1],
                ],
                dtype=np.float64,
            )
            check_sparse_matrix_specific_order(
                self,
                h,
                x_order,
                x_order,
                comparison_h,
                comparison_x_order,
                comparison_x_order,
            )
        else:
            with self.assertRaises(AttributeError):
                h = pyomo_nlp.evaluate_hessian_lag()

    def test_pressure_drop_two_equalities_two_outputs(self):
        self._test_pressure_drop_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputs(), False
        )
        self._test_pressure_drop_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputsWithHessian(), True
        )

    def _test_pressure_drop_two_equalities_two_outputs(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['P1'].value = 80
        m.egb.inputs['P1'].setlb(10)
        m.egb.inputs['P1'].setub(90)
        m.egb.inputs['P3'].value = 70
        m.egb.inputs['P3'].setlb(20)
        m.egb.inputs['P3'].setub(80)
        m.egb.outputs['P2'].value = 75
        m.egb.outputs['P2'].setlb(15)
        m.egb.outputs['P2'].setub(85)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(30)
        m.egb.outputs['Pout'].setub(70)
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)

        self.assertEqual(7, pyomo_nlp.n_primals())
        self.assertEqual(4, pyomo_nlp.n_constraints())
        self.assertEqual(16, pyomo_nlp.nnz_jacobian())
        if hessian_support:
            self.assertEqual(5, pyomo_nlp.nnz_hessian_lag())

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[P1]',
            'egb.inputs[P3]',
            'egb.outputs[P2]',
            'egb.outputs[Pout]',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = ['egb.pdrop1', 'egb.pdrop3', 'egb.P2_con', 'egb.Pout_con']
        c_order = pyomo_nlp.constraint_names()

        xlb = pyomo_nlp.primals_lb()
        comparison_xlb = np.asarray([50, 1, 1, 10, 20, 15, 30], dtype=np.float64)
        check_vectors_specific_order(
            self, xlb, x_order, comparison_xlb, comparison_x_order
        )
        xub = pyomo_nlp.primals_ub()
        comparison_xub = np.asarray([150, 5, 5, 90, 80, 85, 70], dtype=np.float64)
        check_vectors_specific_order(
            self, xub, x_order, comparison_xub, comparison_x_order
        )
        clb = pyomo_nlp.constraints_lb()
        comparison_clb = np.asarray([0, 0, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, clb, c_order, comparison_clb, comparison_c_order
        )
        cub = pyomo_nlp.constraints_ub()
        comparison_cub = np.asarray([0, 0, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, cub, c_order, comparison_cub, comparison_c_order
        )

        xinit = pyomo_nlp.init_primals()
        comparison_xinit = np.asarray([100, 2, 3, 80, 70, 75, 50], dtype=np.float64)
        check_vectors_specific_order(
            self, xinit, x_order, comparison_xinit, comparison_x_order
        )
        duals_init = pyomo_nlp.init_duals()
        comparison_duals_init = np.asarray([0, 0, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, duals_init, c_order, comparison_duals_init, comparison_c_order
        )

        self.assertEqual(7, len(pyomo_nlp.create_new_vector('primals')))
        self.assertEqual(4, len(pyomo_nlp.create_new_vector('constraints')))
        self.assertEqual(4, len(pyomo_nlp.create_new_vector('duals')))
        self.assertEqual(4, len(pyomo_nlp.create_new_vector('eq_constraints')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('ineq_constraints')))
        self.assertEqual(4, len(pyomo_nlp.create_new_vector('duals_eq')))
        self.assertEqual(0, len(pyomo_nlp.create_new_vector('duals_ineq')))

        pyomo_nlp.set_primals(np.asarray([1, 2, 3, 4, 5, 6, 7], dtype=np.float64))
        x = pyomo_nlp.get_primals()
        self.assertTrue(
            np.array_equal(x, np.asarray([1, 2, 3, 4, 5, 6, 7], dtype=np.float64))
        )
        pyomo_nlp.set_primals(pyomo_nlp.init_primals())

        pyomo_nlp.set_duals(np.asarray([42, 10, 11, 12], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(
            np.array_equal(y, np.asarray([42, 10, 11, 12], dtype=np.float64))
        )
        pyomo_nlp.set_duals(np.asarray([21, 5, 6, 7], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(np.array_equal(y, np.asarray([21, 5, 6, 7], dtype=np.float64)))

        fac = pyomo_nlp.get_obj_factor()
        self.assertEqual(fac, 1)
        pyomo_nlp.set_obj_factor(42)
        self.assertEqual(pyomo_nlp.get_obj_factor(), 42)
        pyomo_nlp.set_obj_factor(1)

        f = pyomo_nlp.evaluate_objective()
        self.assertEqual(f, 900)

        gradf = pyomo_nlp.evaluate_grad_objective()
        comparison_gradf = np.asarray([0, 0, 0, 0, 0, 0, 60], dtype=np.float64)
        check_vectors_specific_order(
            self, gradf, x_order, comparison_gradf, comparison_x_order
        )
        c = pyomo_nlp.evaluate_constraints()
        comparison_c = np.asarray([-2, 26, -13, -22], dtype=np.float64)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)
        c = np.zeros(4)
        pyomo_nlp.evaluate_constraints(out=c)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)

        j = pyomo_nlp.evaluate_jacobian()
        comparison_j = np.asarray(
            [
                [-1, 9, 12, 1, 0, 0, 0],
                [0, 18, 24, -1, 1, 0, 0],
                [0, -9, -12, 1, 0, -1, 0],
                [1, -36, -48, 0, 0, 0, -1],
            ]
        )
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        j = 2.0 * j
        pyomo_nlp.evaluate_jacobian(out=j)
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        if hessian_support:
            h = pyomo_nlp.evaluate_hessian_lag()
            self.assertTrue(h.shape == (7, 7))
            comparison_h = np.asarray(
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [
                        0,
                        (2 * 3 * 21) + (4 * 3 * 5) + (-2 * 3 * 6) + (-8 * 3 * 7),
                        (2 * 2 * 21) + (4 * 2 * 5) + (-2 * 2 * 6) + (-8 * 2 * 7),
                        0,
                        0,
                        0,
                        0,
                    ],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 2 * 1],
                ],
                dtype=np.float64,
            )
            check_sparse_matrix_specific_order(
                self,
                h,
                x_order,
                x_order,
                comparison_h,
                comparison_x_order,
                comparison_x_order,
            )
        else:
            with self.assertRaises(AttributeError):
                h = pyomo_nlp.evaluate_hessian_lag()

    def test_external_additional_constraints_vars(self):
        self._test_external_additional_constraints_vars(
            ex_models.PressureDropTwoEqualitiesTwoOutputs(), False
        )
        self._test_external_additional_constraints_vars(
            ex_models.PressureDropTwoEqualitiesTwoOutputsWithHessian(), True
        )

    def _test_external_additional_constraints_vars(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.hin = pyo.Var(bounds=(0, None), initialize=10)
        m.hout = pyo.Var(bounds=(0, None))
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.incon = pyo.Constraint(expr=0 <= m.egb.inputs['Pin'] - 10 * m.hin)
        m.outcon = pyo.Constraint(expr=0 == m.egb.outputs['Pout'] - 10 * m.hout)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['P1'].value = 80
        m.egb.inputs['P1'].setlb(10)
        m.egb.inputs['P1'].setub(90)
        m.egb.inputs['P3'].value = 70
        m.egb.inputs['P3'].setlb(20)
        m.egb.inputs['P3'].setub(80)
        m.egb.outputs['P2'].value = 75
        m.egb.outputs['P2'].setlb(15)
        m.egb.outputs['P2'].setub(85)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(30)
        m.egb.outputs['Pout'].setub(70)
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)

        self.assertEqual(9, pyomo_nlp.n_primals())
        self.assertEqual(6, pyomo_nlp.n_constraints())
        self.assertEqual(20, pyomo_nlp.nnz_jacobian())
        if hessian_support:
            self.assertEqual(5, pyomo_nlp.nnz_hessian_lag())

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[P1]',
            'egb.inputs[P3]',
            'egb.outputs[P2]',
            'egb.outputs[Pout]',
            'hin',
            'hout',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = [
            'egb.pdrop1',
            'egb.pdrop3',
            'egb.P2_con',
            'egb.Pout_con',
            'incon',
            'outcon',
        ]
        c_order = pyomo_nlp.constraint_names()

        xlb = pyomo_nlp.primals_lb()
        comparison_xlb = np.asarray([50, 1, 1, 10, 20, 15, 30, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, xlb, x_order, comparison_xlb, comparison_x_order
        )
        xub = pyomo_nlp.primals_ub()
        comparison_xub = np.asarray(
            [150, 5, 5, 90, 80, 85, 70, np.inf, np.inf], dtype=np.float64
        )
        check_vectors_specific_order(
            self, xub, x_order, comparison_xub, comparison_x_order
        )
        clb = pyomo_nlp.constraints_lb()
        comparison_clb = np.asarray([0, 0, 0, 0, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, clb, c_order, comparison_clb, comparison_c_order
        )
        cub = pyomo_nlp.constraints_ub()
        comparison_cub = np.asarray([0, 0, 0, 0, np.inf, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, cub, c_order, comparison_cub, comparison_c_order
        )

        xinit = pyomo_nlp.init_primals()
        comparison_xinit = np.asarray(
            [100, 2, 3, 80, 70, 75, 50, 10, 0], dtype=np.float64
        )
        check_vectors_specific_order(
            self, xinit, x_order, comparison_xinit, comparison_x_order
        )
        duals_init = pyomo_nlp.init_duals()
        comparison_duals_init = np.asarray([0, 0, 0, 0, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, duals_init, c_order, comparison_duals_init, comparison_c_order
        )

        self.assertEqual(9, len(pyomo_nlp.create_new_vector('primals')))
        self.assertEqual(6, len(pyomo_nlp.create_new_vector('constraints')))
        self.assertEqual(6, len(pyomo_nlp.create_new_vector('duals')))
        self.assertEqual(5, len(pyomo_nlp.create_new_vector('eq_constraints')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('ineq_constraints')))
        self.assertEqual(5, len(pyomo_nlp.create_new_vector('duals_eq')))
        self.assertEqual(1, len(pyomo_nlp.create_new_vector('duals_ineq')))

        pyomo_nlp.set_primals(np.asarray([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.float64))
        x = pyomo_nlp.get_primals()
        self.assertTrue(
            np.array_equal(x, np.asarray([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.float64))
        )
        pyomo_nlp.set_primals(pyomo_nlp.init_primals())

        pyomo_nlp.set_duals(np.asarray([42, 10, 11, 12, 13, 14], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(
            np.array_equal(y, np.asarray([42, 10, 11, 12, 13, 14], dtype=np.float64))
        )
        pyomo_nlp.set_duals(np.asarray([0, 0, 21, 5, 6, 7], dtype=np.float64))
        y = pyomo_nlp.get_duals()
        self.assertTrue(
            np.array_equal(y, np.asarray([0, 0, 21, 5, 6, 7], dtype=np.float64))
        )

        fac = pyomo_nlp.get_obj_factor()
        self.assertEqual(fac, 1)
        pyomo_nlp.set_obj_factor(42)
        self.assertEqual(pyomo_nlp.get_obj_factor(), 42)
        pyomo_nlp.set_obj_factor(1)

        f = pyomo_nlp.evaluate_objective()
        self.assertEqual(f, 900)

        gradf = pyomo_nlp.evaluate_grad_objective()
        comparison_gradf = np.asarray([0, 0, 0, 0, 0, 0, 60, 0, 0], dtype=np.float64)
        check_vectors_specific_order(
            self, gradf, x_order, comparison_gradf, comparison_x_order
        )
        c = pyomo_nlp.evaluate_constraints()
        comparison_c = np.asarray([-2, 26, -13, -22, 0, 50], dtype=np.float64)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)
        c = np.zeros(6)
        pyomo_nlp.evaluate_constraints(out=c)
        check_vectors_specific_order(self, c, c_order, comparison_c, comparison_c_order)

        j = pyomo_nlp.evaluate_jacobian()
        comparison_j = np.asarray(
            [
                [-1, 9, 12, 1, 0, 0, 0, 0, 0],
                [0, 18, 24, -1, 1, 0, 0, 0, 0],
                [0, -9, -12, 1, 0, -1, 0, 0, 0],
                [1, -36, -48, 0, 0, 0, -1, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, -10, 0],
                [0, 0, 0, 0, 0, 0, 1, 0, -10],
            ]
        )

        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        j = 2.0 * j
        pyomo_nlp.evaluate_jacobian(out=j)
        check_sparse_matrix_specific_order(
            self,
            j,
            c_order,
            x_order,
            comparison_j,
            comparison_c_order,
            comparison_x_order,
        )

        if hessian_support:
            h = pyomo_nlp.evaluate_hessian_lag()
            self.assertTrue(h.shape == (9, 9))
            comparison_h = np.asarray(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [
                        0,
                        (2 * 3 * 21) + (4 * 3 * 5) + (-2 * 3 * 6) + (-8 * 3 * 7),
                        (2 * 2 * 21) + (4 * 2 * 5) + (-2 * 2 * 6) + (-8 * 2 * 7),
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                    ],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 2 * 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                ],
                dtype=np.float64,
            )
            check_sparse_matrix_specific_order(
                self,
                h,
                x_order,
                x_order,
                comparison_h,
                comparison_x_order,
                comparison_x_order,
            )
        else:
            with self.assertRaises(AttributeError):
                h = pyomo_nlp.evaluate_hessian_lag()

    @unittest.skipIf(not cyipopt_available, "CyIpopt needed to run tests with solve")
    def test_external_greybox_solve(self):
        self._test_external_greybox_solve(
            ex_models.PressureDropTwoEqualitiesTwoOutputs(), False
        )
        self._test_external_greybox_solve(
            ex_models.PressureDropTwoEqualitiesTwoOutputsWithHessian(), True
        )

    def _test_external_greybox_solve(self, ex_model, hessian_support):
        m = pyo.ConcreteModel()
        m.mu = pyo.Var(bounds=(0, None), initialize=1)
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)
        m.ccon = pyo.Constraint(
            expr=m.egb.inputs['c'] == 128 / (3.14 * 1e-4) * m.mu * m.egb.inputs['F']
        )
        m.pcon = pyo.Constraint(expr=m.egb.inputs['Pin'] - m.egb.outputs['Pout'] <= 72)
        m.pincon = pyo.Constraint(expr=m.egb.inputs['Pin'] == 100.0)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['P1'].value = 80
        m.egb.inputs['P1'].setlb(10)
        m.egb.inputs['P1'].setub(90)
        m.egb.inputs['P3'].value = 70
        m.egb.inputs['P3'].setlb(20)
        m.egb.inputs['P3'].setub(80)
        m.egb.outputs['P2'].value = 75
        m.egb.outputs['P2'].setlb(15)
        m.egb.outputs['P2'].setub(85)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(10)
        m.egb.outputs['Pout'].setub(70)
        m.obj = pyo.Objective(
            expr=(m.egb.outputs['Pout'] - 20) ** 2 + (m.egb.inputs['F'] - 3) ** 2
        )

        solver = pyo.SolverFactory('cyipopt')

        if not hessian_support:
            solver.config.options = {'hessian_approximation': 'limited-memory'}
        status = solver.solve(m, tee=False)

        self.assertAlmostEqual(pyo.value(m.egb.inputs['F']), 3.0, places=3)
        self.assertAlmostEqual(pyo.value(m.mu), 1.63542e-6, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.outputs['Pout']), 28.0, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.inputs['Pin']), 100.0, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.inputs['c']), 2.0, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.inputs['P1']), 82.0, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.inputs['P3']), 46.0, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.outputs['P2']), 64.0, places=3)
        self.assertAlmostEqual(pyo.value(m.egb.inputs['F']), 3.0, places=3)

    def create_model_two_equalities_two_outputs(self, external_model):
        m = pyo.ConcreteModel()
        m.hin = pyo.Var(bounds=(0, None), initialize=10)
        m.hout = pyo.Var(bounds=(0, None))
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(external_model)
        m.incon = pyo.Constraint(expr=0 <= m.egb.inputs['Pin'] - 10 * m.hin)
        m.outcon = pyo.Constraint(expr=0 == m.egb.outputs['Pout'] - 10 * m.hout)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['P1'].value = 80
        m.egb.inputs['P1'].setlb(10)
        m.egb.inputs['P1'].setub(90)
        m.egb.inputs['P3'].value = 70
        m.egb.inputs['P3'].setlb(20)
        m.egb.inputs['P3'].setub(80)
        m.egb.outputs['P2'].value = 75
        m.egb.outputs['P2'].setlb(15)
        m.egb.outputs['P2'].setub(85)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(30)
        m.egb.outputs['Pout'].setub(70)
        return m

    def test_scaling_all_missing(self):
        m = self.create_model_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputs()
        )
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)
        fs = pyomo_nlp.get_obj_scaling()
        xs = pyomo_nlp.get_primals_scaling()
        cs = pyomo_nlp.get_constraints_scaling()
        self.assertIsNone(fs)
        self.assertIsNone(xs)
        self.assertIsNone(cs)

    def test_scaling_pyomo_model_only(self):
        m = self.create_model_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputs()
        )
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        # m.scaling_factor[m.obj] = 0.1 # scale the objective
        m.scaling_factor[m.egb.inputs['Pin']] = 1.1  # scale the variable
        m.scaling_factor[m.egb.inputs['c']] = 1.2  # scale the variable
        m.scaling_factor[m.egb.inputs['F']] = 1.3  # scale the variable
        # m.scaling_factor[m.egb.inputs['P1']] = 1.4 # scale the variable
        m.scaling_factor[m.egb.inputs['P3']] = 1.5  # scale the variable
        m.scaling_factor[m.egb.outputs['P2']] = 1.6  # scale the variable
        m.scaling_factor[m.egb.outputs['Pout']] = 1.7  # scale the variable
        # m.scaling_factor[m.hin] = 1.8
        m.scaling_factor[m.hout] = 1.9
        # m.scaling_factor[m.incon] = 2.1
        m.scaling_factor[m.outcon] = 2.2
        pyomo_nlp = PyomoGreyBoxNLP(m)

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[P1]',
            'egb.inputs[P3]',
            'egb.outputs[P2]',
            'egb.outputs[Pout]',
            'hin',
            'hout',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = [
            'egb.pdrop1',
            'egb.pdrop3',
            'egb.P2_con',
            'egb.Pout_con',
            'incon',
            'outcon',
        ]
        c_order = pyomo_nlp.constraint_names()

        fs = pyomo_nlp.get_obj_scaling()
        self.assertEqual(fs, 1.0)

        xs = pyomo_nlp.get_primals_scaling()
        comparison_xs = np.asarray(
            [1.1, 1.2, 1.3, 1.0, 1.5, 1.6, 1.7, 1.0, 1.9], dtype=np.float64
        )
        check_vectors_specific_order(
            self, xs, x_order, comparison_xs, comparison_x_order
        )

        cs = pyomo_nlp.get_constraints_scaling()
        comparison_cs = np.asarray([1, 1, 1, 1, 1, 2.2], dtype=np.float64)
        check_vectors_specific_order(
            self, cs, c_order, comparison_cs, comparison_c_order
        )

    def test_scaling_greybox_only(self):
        m = self.create_model_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputsScaleBoth()
        )
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[P1]',
            'egb.inputs[P3]',
            'egb.outputs[P2]',
            'egb.outputs[Pout]',
            'hin',
            'hout',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = [
            'egb.pdrop1',
            'egb.pdrop3',
            'egb.P2_con',
            'egb.Pout_con',
            'incon',
            'outcon',
        ]
        c_order = pyomo_nlp.constraint_names()

        fs = pyomo_nlp.get_obj_scaling()
        self.assertEqual(fs, 1.0)

        xs = pyomo_nlp.get_primals_scaling()
        comparison_xs = np.asarray([1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=np.float64)
        check_vectors_specific_order(
            self, xs, x_order, comparison_xs, comparison_x_order
        )

        cs = pyomo_nlp.get_constraints_scaling()
        comparison_cs = np.asarray([3.1, 3.2, 4.1, 4.2, 1, 1], dtype=np.float64)
        check_vectors_specific_order(
            self, cs, c_order, comparison_cs, comparison_c_order
        )

        m = self.create_model_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputsScaleEqualities()
        )
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)
        cs = pyomo_nlp.get_constraints_scaling()
        comparison_cs = np.asarray([3.1, 3.2, 1, 1, 1, 1], dtype=np.float64)
        check_vectors_specific_order(
            self, cs, c_order, comparison_cs, comparison_c_order
        )

        m = self.create_model_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputsScaleOutputs()
        )
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        pyomo_nlp = PyomoGreyBoxNLP(m)
        cs = pyomo_nlp.get_constraints_scaling()
        comparison_cs = np.asarray([1, 1, 4.1, 4.2, 1, 1], dtype=np.float64)
        check_vectors_specific_order(
            self, cs, c_order, comparison_cs, comparison_c_order
        )

    def test_scaling_pyomo_model_and_greybox(self):
        m = self.create_model_two_equalities_two_outputs(
            ex_models.PressureDropTwoEqualitiesTwoOutputsScaleBoth()
        )
        m.obj = pyo.Objective(expr=(m.egb.outputs['Pout'] - 20) ** 2)
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        # m.scaling_factor[m.obj] = 0.1 # scale the objective
        m.scaling_factor[m.egb.inputs['Pin']] = 1.1  # scale the variable
        m.scaling_factor[m.egb.inputs['c']] = 1.2  # scale the variable
        m.scaling_factor[m.egb.inputs['F']] = 1.3  # scale the variable
        # m.scaling_factor[m.egb.inputs['P1']] = 1.4 # scale the variable
        m.scaling_factor[m.egb.inputs['P3']] = 1.5  # scale the variable
        m.scaling_factor[m.egb.outputs['P2']] = 1.6  # scale the variable
        m.scaling_factor[m.egb.outputs['Pout']] = 1.7  # scale the variable
        # m.scaling_factor[m.hin] = 1.8
        m.scaling_factor[m.hout] = 1.9
        # m.scaling_factor[m.incon] = 2.1
        m.scaling_factor[m.outcon] = 2.2
        pyomo_nlp = PyomoGreyBoxNLP(m)

        comparison_x_order = [
            'egb.inputs[Pin]',
            'egb.inputs[c]',
            'egb.inputs[F]',
            'egb.inputs[P1]',
            'egb.inputs[P3]',
            'egb.outputs[P2]',
            'egb.outputs[Pout]',
            'hin',
            'hout',
        ]
        x_order = pyomo_nlp.variable_names()
        comparison_c_order = [
            'egb.pdrop1',
            'egb.pdrop3',
            'egb.P2_con',
            'egb.Pout_con',
            'incon',
            'outcon',
        ]
        c_order = pyomo_nlp.constraint_names()

        fs = pyomo_nlp.get_obj_scaling()
        self.assertEqual(fs, 1.0)

        xs = pyomo_nlp.get_primals_scaling()
        comparison_xs = np.asarray(
            [1.1, 1.2, 1.3, 1.0, 1.5, 1.6, 1.7, 1.0, 1.9], dtype=np.float64
        )
        check_vectors_specific_order(
            self, xs, x_order, comparison_xs, comparison_x_order
        )

        cs = pyomo_nlp.get_constraints_scaling()
        comparison_cs = np.asarray([3.1, 3.2, 4.1, 4.2, 1, 2.2], dtype=np.float64)
        check_vectors_specific_order(
            self, cs, c_order, comparison_cs, comparison_c_order
        )

    @unittest.skipIf(not cyipopt_available, "CyIpopt needed to run tests with solve")
    def test_external_greybox_solve_scaling(self):
        m = pyo.ConcreteModel()
        m.mu = pyo.Var(bounds=(0, None), initialize=1)
        m.egb = ExternalGreyBoxBlock()
        m.egb.set_external_model(
            ex_models.PressureDropTwoEqualitiesTwoOutputsScaleBoth()
        )
        m.ccon = pyo.Constraint(
            expr=m.egb.inputs['c'] == 128 / (3.14 * 1e-4) * m.mu * m.egb.inputs['F']
        )
        m.pcon = pyo.Constraint(expr=m.egb.inputs['Pin'] - m.egb.outputs['Pout'] <= 72)
        m.pincon = pyo.Constraint(expr=m.egb.inputs['Pin'] == 100.0)
        m.egb.inputs['Pin'].value = 100
        m.egb.inputs['Pin'].setlb(50)
        m.egb.inputs['Pin'].setub(150)
        m.egb.inputs['c'].value = 2
        m.egb.inputs['c'].setlb(1)
        m.egb.inputs['c'].setub(5)
        m.egb.inputs['F'].value = 3
        m.egb.inputs['F'].setlb(1)
        m.egb.inputs['F'].setub(5)
        m.egb.inputs['P1'].value = 80
        m.egb.inputs['P1'].setlb(10)
        m.egb.inputs['P1'].setub(90)
        m.egb.inputs['P3'].value = 70
        m.egb.inputs['P3'].setlb(20)
        m.egb.inputs['P3'].setub(80)
        m.egb.outputs['P2'].value = 75
        m.egb.outputs['P2'].setlb(15)
        m.egb.outputs['P2'].setub(85)
        m.egb.outputs['Pout'].value = 50
        m.egb.outputs['Pout'].setlb(10)
        m.egb.outputs['Pout'].setub(70)
        m.obj = pyo.Objective(
            expr=(m.egb.outputs['Pout'] - 20) ** 2 + (m.egb.inputs['F'] - 3) ** 2
        )

        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.scaling_factor[m.obj] = 0.1  # scale the objective
        m.scaling_factor[m.egb.inputs['Pin']] = 1.1  # scale the variable
        m.scaling_factor[m.egb.inputs['c']] = 1.2  # scale the variable
        m.scaling_factor[m.egb.inputs['F']] = 1.3  # scale the variable
        # m.scaling_factor[m.egb.inputs['P1']] = 1.4 # scale the variable
        m.scaling_factor[m.egb.inputs['P3']] = 1.5  # scale the variable
        m.scaling_factor[m.egb.outputs['P2']] = 1.6  # scale the variable
        m.scaling_factor[m.egb.outputs['Pout']] = 1.7  # scale the variable
        m.scaling_factor[m.mu] = 1.9
        m.scaling_factor[m.pincon] = 2.2

        with TempfileManager.new_context() as temp:
            logfile = temp.create_tempfile('_cyipopt-external-greybox-scaling.log')
            solver = pyo.SolverFactory('cyipopt')
            solver.config.options = {
                'hessian_approximation': 'limited-memory',
                'nlp_scaling_method': 'user-scaling',
                'output_file': logfile,
                'file_print_level': 10,
                'max_iter': 0,
            }
            status = solver.solve(m, tee=False)

            with open(logfile, 'r') as fd:
                solver_trace = fd.read()

        self.assertIn('nlp_scaling_method = user-scaling', solver_trace)
        self.assertIn(f'output_file = {logfile}', solver_trace)
        self.assertIn('objective scaling factor = 0.1', solver_trace)
        self.assertIn('x scaling provided', solver_trace)
        self.assertIn('c scaling provided', solver_trace)
        self.assertIn('d scaling provided', solver_trace)
        # x order: ['egb.inputs[F]', 'mu', 'egb.outputs[Pout]', 'egb.inputs[Pin]', 'egb.inputs[c]', 'egb.inputs[P1]', 'egb.inputs[P3]', 'egb.outputs[P2]']
        # c order: ['ccon', 'pcon', 'pincon', 'egb.pdrop1', 'egb.pdrop3', 'egb.P2_con', 'egb.Pout_con']
        self.assertIn('DenseVector "x scaling vector" with 8 elements:', solver_trace)
        self.assertIn(
            'x scaling vector[    1]= 1.3000000000000000e+00', solver_trace
        )  # F
        self.assertIn(
            'x scaling vector[    8]= 1.8999999999999999e+00', solver_trace
        )  # mu
        self.assertIn(
            'x scaling vector[    7]= 1.7000000000000000e+00', solver_trace
        )  # Pout
        self.assertIn(
            'x scaling vector[    4]= 1.1000000000000001e+00', solver_trace
        )  # Pin
        self.assertIn(
            'x scaling vector[    5]= 1.2000000000000000e+00', solver_trace
        )  # c
        self.assertIn(
            'x scaling vector[    2]= 1.0000000000000000e+00', solver_trace
        )  # P1
        self.assertIn(
            'x scaling vector[    3]= 1.5000000000000000e+00', solver_trace
        )  # P3
        self.assertIn(
            'x scaling vector[    6]= 1.6000000000000001e+00', solver_trace
        )  # P2
        self.assertIn('DenseVector "c scaling vector" with 6 elements:', solver_trace)
        self.assertIn(
            'c scaling vector[    1]= 1.0000000000000000e+00', solver_trace
        )  # ccon
        self.assertIn(
            'c scaling vector[    2]= 2.2000000000000002e+00', solver_trace
        )  # pincon
        self.assertIn(
            'c scaling vector[    3]= 3.1000000000000001e+00', solver_trace
        )  # pdrop1
        self.assertIn(
            'c scaling vector[    4]= 3.2000000000000002e+00', solver_trace
        )  # pdrop3
        self.assertIn(
            'c scaling vector[    5]= 4.0999999999999996e+00', solver_trace
        )  # P2_con
        self.assertIn(
            'c scaling vector[    6]= 4.2000000000000002e+00', solver_trace
        )  # Pout_con
        self.assertIn('DenseVector "d scaling vector" with 1 elements:', solver_trace)
        self.assertIn(
            'd scaling vector[    1]= 1.0000000000000000e+00', solver_trace
        )  # pcon


if __name__ == '__main__':
    TestPyomoGreyBoxNLP().test_external_greybox_solve()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

import pyomo.common.unittest as unittest

import pyomo.environ as pyo

from pyomo.common.dependencies import numpy as np, scipy_available, numpy_available
from pyomo.common.log import LoggingIntercept
from pyomo.repn.plugins.standard_form import LinearStandardFormCompiler

import pyomo.core.base.constraint as constraint
import pyomo.core.base.objective as objective

for sol in ['glpk', 'cbc', 'gurobi', 'cplex', 'xpress']:
    linear_solver = pyo.SolverFactory(sol)
    if linear_solver.available(exception_flag=False):
        break
else:
    linear_solver = None


@unittest.skipUnless(
    scipy_available & numpy_available, "standard_form requires scipy and numpy"
)
class TestLinearStandardFormCompiler(unittest.TestCase):
    def push_templatization(self, mode):
        self.templatize_stack.append(
            (constraint.TEMPLATIZE_CONSTRAINTS, objective.TEMPLATIZE_OBJECTIVES)
        )
        constraint.TEMPLATIZE_CONSTRAINTS = mode
        objective.TEMPLATIZE_OBJECTIVES = mode

    def pop_templatization(self):
        constraint.TEMPLATIZE_CONSTRAINTS, objective.TEMPLATIZE_OBJECTIVES = (
            self.templatize_stack.pop()
        )

    def setUp(self):
        self.templatize_stack = []

    def test_linear_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])
        m.c = pyo.Constraint(expr=m.x + 2 * m.y[1] >= 3)
        m.d = pyo.Constraint(expr=m.y[1] + 4 * m.y[3] <= 5)

        repn = LinearStandardFormCompiler().write(m)

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        self.assertTrue(np.all(repn.A == np.array([[-1, -2, 0], [0, 1, 4]])))
        self.assertTrue(np.all(repn.rhs == np.array([-3, 5])))
        self.assertEqual(repn.rows, [(m.c, -1), (m.d, 1)])
        self.assertEqual(repn.columns, [m.x, m.y[1], m.y[3]])

    def test_almost_dense_linear_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])
        m.c = pyo.Constraint(expr=m.x + 2 * m.y[1] + 4 * m.y[3] >= 10)
        m.d = pyo.Constraint(expr=5 * m.x + 6 * m.y[1] + 8 * m.y[3] <= 20)

        repn = LinearStandardFormCompiler().write(m)

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        self.assertTrue(np.all(repn.A == np.array([[-1, -2, -4], [5, 6, 8]])))
        self.assertTrue(np.all(repn.rhs == np.array([-10, 20])))
        self.assertEqual(repn.rows, [(m.c, -1), (m.d, 1)])
        self.assertEqual(repn.columns, [m.x, m.y[1], m.y[3]])

    def test_linear_model_row_col_order(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])
        m.c = pyo.Constraint(expr=m.x + 2 * m.y[1] >= 3)
        m.d = pyo.Constraint(expr=m.y[1] + 4 * m.y[3] <= 5)

        repn = LinearStandardFormCompiler().write(
            m, column_order=[m.y[3], m.y[2], m.x, m.y[1]], row_order=[m.d, m.c]
        )

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        self.assertTrue(np.all(repn.A == np.array([[4, 0, 1], [0, -1, -2]])))
        self.assertTrue(np.all(repn.rhs == np.array([5, -3])))
        self.assertEqual(repn.rows, [(m.d, 1), (m.c, -1)])
        self.assertEqual(repn.columns, [m.y[3], m.x, m.y[1]])

    def test_linear_model_fixed_vars(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3, 4, 5])
        m.o = pyo.Objective(expr=5 * m.x + 3 * m.y[5] + 1)
        m.c = pyo.Constraint(expr=m.x + 2 * m.y[1] >= 3)
        m.d = pyo.Constraint(expr=m.y[1] + 4 * m.y[3] <= 5)

        m.y[3].fix(10)
        m.y[4].fix(100)
        m.y[5].fix(1000)

        repn = LinearStandardFormCompiler().write(m)

        self.assertTrue(np.all(repn.c == np.array([5, 0])))
        self.assertEqual(repn.c_offset, 3001)
        self.assertTrue(np.all(repn.A == np.array([[-1, -2], [0, 1]])))
        self.assertTrue(np.all(repn.rhs == np.array([-3, -35])))
        self.assertEqual(repn.rows, [(m.c, -1), (m.d, 1)])
        self.assertEqual(repn.columns, [m.x, m.y[1]])

    def test_suffix_warning(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var([1, 2, 3])
        m.c = pyo.Constraint(expr=m.x + 2 * m.y[1] >= 3)
        m.d = pyo.Constraint(expr=m.y[1] + 4 * m.y[3] <= 5)

        m.dual = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.b = pyo.Block()
        m.b.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT_EXPORT)

        with LoggingIntercept() as LOG:
            repn = LinearStandardFormCompiler().write(m)
        self.assertEqual(LOG.getvalue(), "")

        m.dual[m.c] = 5
        with LoggingIntercept() as LOG:
            repn = LinearStandardFormCompiler().write(m)
        self.assertEqual(
            LOG.getvalue(),
            "EXPORT Suffix 'dual' found on 1 block:\n"
            "    dual\n"
            "Standard Form compiler ignores export suffixes.  Skipping.\n",
        )

        m.b.dual[m.d] = 1
        with LoggingIntercept() as LOG:
            repn = LinearStandardFormCompiler().write(m)
        self.assertEqual(
            LOG.getvalue(),
            "EXPORT Suffix 'dual' found on 2 blocks:\n"
            "    dual\n"
            "    b.dual\n"
            "Standard Form compiler ignores export suffixes.  Skipping.\n",
        )

    def _verify_solution(self, soln, repn, eq):
        # clear out any old solution
        for v, val in soln:
            v.value = None
        for v in repn.x:
            v.value = None

        x = np.array(repn.x, dtype=object)
        ax = repn.A.todense() @ x

        def c_rule(m, i):
            if eq:
                return ax[i] == repn.b[i]
            else:
                return ax[i] <= repn.b[i]

        try:
            self.push_templatization(False)
            test_model = pyo.ConcreteModel()
            test_model.o = pyo.Objective(expr=repn.c[[1], :].todense()[0] @ x)
            test_model.c = pyo.Constraint(range(len(repn.b)), rule=c_rule)
        finally:
            self.pop_templatization()
        linear_solver.solve(test_model, tee=True)

        # Propagate any solution back to the original variables
        for v, expr in repn.eliminated_vars:
            v.value = pyo.value(expr)
        self.assertEqual(*zip(*((v.value, val) for v, val in soln)))

    @unittest.skipIf(
        linear_solver is None, 'verifying results requires a linear solver'
    )
    def test_alternative_forms(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var(
            [0, 1, 3], bounds=lambda m, i: (-1 * (i % 2) * 5, 10 - 12 * (i // 2))
        )
        m.c = pyo.Constraint(expr=m.x + 2 * m.y[1] >= 3)
        m.d = pyo.Constraint(expr=m.y[1] + 4 * m.y[3] <= 5)
        m.e = pyo.Constraint(expr=pyo.inequality(-2, m.y[0] + 1 + 6 * m.y[1], 7))
        m.f = pyo.Constraint(expr=m.x + m.y[0] + 2 == 10)
        m.o = pyo.Objective([1, 3], rule=lambda m, i: m.x + i * 5 * m.y[i])
        m.o[1].sense = pyo.maximize

        col_order = [m.x, m.y[0], m.y[1], m.y[3]]

        m.o[1].deactivate()
        linear_solver.solve(m)
        m.o[1].activate()
        soln = [(v, v.value) for v in col_order]

        repn = LinearStandardFormCompiler().write(m, column_order=col_order)

        self.assertEqual(
            repn.rows, [(m.c, -1), (m.d, 1), (m.e, 1), (m.e, -1), (m.f, 1), (m.f, -1)]
        )
        self.assertEqual(repn.x, [m.x, m.y[0], m.y[1], m.y[3]])
        ref = np.array(
            [
                [-1, 0, -2, 0],
                [0, 0, 1, 4],
                [0, 1, 6, 0],
                [0, -1, -6, 0],
                [1, 1, 0, 0],
                [-1, -1, 0, 0],
            ]
        )
        self.assertTrue(np.all(repn.A == ref))
        self.assertTrue(np.all(repn.b == np.array([-3, 5, 6, 3, 8, -8])))
        self.assertTrue(np.all(repn.c == np.array([[-1, 0, -5, 0], [1, 0, 0, 15]])))
        self._verify_solution(soln, repn, False)

        repn = LinearStandardFormCompiler().write(
            m, nonnegative_vars=True, column_order=col_order
        )

        self.assertEqual(
            repn.rows, [(m.c, -1), (m.d, 1), (m.e, 1), (m.e, -1), (m.f, 1), (m.f, -1)]
        )
        self.assertEqual(
            list(map(str, repn.x)),
            ['_neg_0', '_pos_0', 'y[0]', '_neg_2', '_pos_2', '_neg_3'],
        )
        ref = np.array(
            [
                [1, -1, 0, 2, -2, 0],
                [0, 0, 0, -1, 1, -4],
                [0, 0, 1, -6, 6, 0],
                [0, 0, -1, 6, -6, 0],
                [-1, 1, 1, 0, 0, 0],
                [1, -1, -1, 0, 0, 0],
            ]
        )
        self.assertTrue(np.all(repn.A == ref))
        self.assertTrue(np.all(repn.b == np.array([-3, 5, 6, 3, 8, -8])))
        self.assertTrue(
            np.all(repn.c == np.array([[1, -1, 0, 5, -5, 0], [-1, 1, 0, 0, 0, -15]]))
        )
        self._verify_solution(soln, repn, False)

        repn = LinearStandardFormCompiler().write(
            m, slack_form=True, column_order=col_order
        )

        self.assertEqual(repn.rows, [(m.c, 1), (m.d, 1), (m.e, 1), (m.f, 1)])
        self.assertEqual(
            list(map(str, repn.x)),
            ['x', 'y[0]', 'y[1]', 'y[3]', '_slack_0', '_slack_1', '_slack_2'],
        )
        self.assertEqual(
            list(v.bounds for v in repn.x),
            [(None, None), (0, 10), (-5, 10), (-5, -2), (None, 0), (0, None), (-9, 0)],
        )
        ref = np.array(
            [
                [1, 0, 2, 0, 1, 0, 0],
                [0, 0, 1, 4, 0, 1, 0],
                [0, 1, 6, 0, 0, 0, 1],
                [1, 1, 0, 0, 0, 0, 0],
            ]
        )
        self.assertTrue(np.all(repn.A == ref))
        self.assertTrue(np.all(repn.b == np.array([3, 5, -3, 8])))
        self.assertTrue(
            np.all(
                repn.c == np.array([[-1, 0, -5, 0, 0, 0, 0], [1, 0, 0, 15, 0, 0, 0]])
            )
        )
        self._verify_solution(soln, repn, True)

        repn = LinearStandardFormCompiler().write(
            m, mixed_form=True, column_order=col_order
        )

        self.assertEqual(
            repn.rows, [(m.c, -1), (m.d, 1), (m.e, 1), (m.e, -1), (m.f, 0)]
        )
        self.assertEqual(list(map(str, repn.x)), ['x', 'y[0]', 'y[1]', 'y[3]'])
        self.assertEqual(
            list(v.bounds for v in repn.x), [(None, None), (0, 10), (-5, 10), (-5, -2)]
        )
        ref = np.array(
            [[1, 0, 2, 0], [0, 0, 1, 4], [0, 1, 6, 0], [0, 1, 6, 0], [1, 1, 0, 0]]
        )
        self.assertTrue(np.all(repn.A == ref))
        self.assertTrue(np.all(repn.b == np.array([3, 5, 6, -3, 8])))
        self.assertTrue(np.all(repn.c == np.array([[-1, 0, -5, 0], [1, 0, 0, 15]])))
        # Note that the mixed_form solution is a mix of inequality and
        # equality constraints, so we cannot (easily) reuse the
        # _verify_solutions helper (as in the above cases):
        # self._verify_solution(soln, repn, False)

        repn = LinearStandardFormCompiler().write(
            m, slack_form=True, nonnegative_vars=True, column_order=col_order
        )

        self.assertEqual(repn.rows, [(m.c, 1), (m.d, 1), (m.e, 1), (m.f, 1)])
        self.assertEqual(
            list(map(str, repn.x)),
            [
                '_neg_0',
                '_pos_0',
                'y[0]',
                '_neg_2',
                '_pos_2',
                '_neg_3',
                '_neg_4',
                '_slack_1',
                '_neg_6',
            ],
        )
        self.assertEqual(
            list(v.bounds for v in repn.x),
            [
                (0, None),
                (0, None),
                (0, 10),
                (0, 5),
                (0, 10),
                (2, 5),
                (0, None),
                (0, None),
                (0, 9),
            ],
        )
        ref = np.array(
            [
                [-1, 1, 0, -2, 2, 0, -1, 0, 0],
                [0, 0, 0, -1, 1, -4, 0, 1, 0],
                [0, 0, 1, -6, 6, 0, 0, 0, -1],
                [-1, 1, 1, 0, 0, 0, 0, 0, 0],
            ]
        )
        self.assertTrue(np.all(repn.A == ref))
        self.assertTrue(np.all(repn.b == np.array([3, 5, -3, 8])))
        ref = np.array([[1, -1, 0, 5, -5, 0, 0, 0, 0], [-1, 1, 0, 0, 0, -15, 0, 0, 0]])
        self.assertTrue(np.all(repn.c == ref))
        self._verify_solution(soln, repn, True)


class TestTemplatedLinearStandardFormCompiler(TestLinearStandardFormCompiler):
    def setUp(self):
        super().setUp()
        self.push_templatization(True)

    def tearDown(self):
        self.pop_templatization()
